/***************************************************************************
 * File:          channel.c
 * Description:   issue command / interrupt handling for IDE layer
 * Copyright (c)  2000-2003 HighPoint Technologies, Inc. All rights reserved
 ***************************************************************************/

#include "global.h"



/******************************************************************
 * Physical disk command done. Start Next command.
 *******************************************************************/
static void HPTLIBAPI PhysDiskDone(_VBUS_ARG PCommand pCmd)
{
	PDevice pCurrent = &pCmd->pVDevice->u.disk, pDev;
	PChannel pChan = pCurrent->pChannel;
	PVBus pVBus = pChan->pChipInstance->pVBus;

	HPT_ASSERT(pChan->CurrentCmd==pCmd);

	pChan->pChWorkDev = 0;
	pChan->CurrentCmd = 0;
	pVBus->working_devs--;
	
#ifdef CHECK_HARDWARE_TIMEOUT
	if (pChan->Timer) {
		DelTimerCall(_VBUS_P pChan->Timer);
		pChan->Timer = 0;
	}
#endif

	pCmd->pfnCompletion = pCmd->uScratch.disk.pfnOrgDone;
	pCmd->uScratch.disk.pfnOrgDone = 0;
	

	if (pCmd->Result==RETURN_RETRY) {
		fChannelSendCommand(_VBUS_P pCurrent, pCmd);
		return;
	}


	pCmd->pfnCompletion(_VBUS_P pCmd); /* CANNOT use pCmd after this call! */
	
	if (pVBus->serial_mode) {
		if (pVBus->working_devs==0) {
			int i = pVBus->next_active;
			do {
				
				PChannel pChan2;
				i = (i+1) % (MAX_CHANNEL_PER_CHIP*2*pVBus->nInstances);
				pChan2 = pVBus->pChipInstance[i/(MAX_CHANNEL_PER_CHIP*2)]->pChannel[(i/2)%MAX_CHANNEL_PER_CHIP];
				if (pChan2) {
					pDev = pChan2->pDevice[i&1];
				if (pDev && (pCmd=GetCommandFromQueue(pDev))) {
					pVBus->next_active = i;
					fChannelSendCommand(_VBUS_P pDev, pCmd);
					return;
				}
			}
			}
			while (i!=pVBus->next_active);
			pVBus->serial_mode = 0;
		}
	}
	else if (pChan->pChWorkDev==0) {
#ifdef CHECK_CLOCK_SWITCH
		PChipInstance pChip = pChan->pChipInstance;
		/* check another channel */
		if (pChip->ftbl.pfnSwitchClock) {
			PChannel pChan2 = (pChip->pChannel[0]==pChan)? pChip->pChannel[1] : pChip->pChannel[0];
			if (pChan2->CurrentCmd)
				goto check_this;
			else if ((pDev = pChan2->pDevice[pChan2->next_active]) &&
					(pCmd=GetCommandFromQueue(pDev)))
			{
				pChan2->next_active ^= 1;
				fChannelSendCommand(_VBUS_P pDev, pCmd);
			}
			else if ((pDev = pChan2->pDevice[pChan2->next_active^1]) &&
					(pCmd=GetCommandFromQueue(pDev)))
			{
				fChannelSendCommand(_VBUS_P pDev, pCmd);
			}
check_this:
			pDev = (pCurrent==pChan->pDevice[0])? pChan->pDevice[1] : pChan->pDevice[0];
			if (pDev && (pCmd = PickCommandFromQueue(pDev))) {
				if ((pCmd->cf_data_in==0 && pCmd->cf_data_out==0) ||
					pChip->ftbl.pfnClockCapable(pChan, pCmd->cf_data_in))
				{
					GetCommandFromQueue(pDev);
					fChannelSendCommand(_VBUS_P pDev, pCmd);
				}
				/* don't check pCurrent. Otherwise pDev may get starved */
				return;
			}
			/* continue my work */
			if ((pCmd=PickCommandFromQueue(pCurrent)) &&
				((pCmd->cf_data_in==0 && pCmd->cf_data_out==0) ||
				pChip->ftbl.pfnClockCapable(pChan, pCmd->cf_data_in)))
			{
				GetCommandFromQueue(pCurrent);
				fChannelSendCommand(_VBUS_P pCurrent, pCmd);
			}
			return;
		}
#endif
		/* check another drive on the channel first */
		pDev = (pCurrent==pChan->pDevice[0])? pChan->pDevice[1] : pChan->pDevice[0];
		if (pDev) {
			if ((pCmd=GetCommandFromQueue(pDev))) {
				fChannelSendCommand(_VBUS_P pDev, pCmd);
				return;
			}
		}
		/* continue my work */
		if ((pCmd=GetCommandFromQueue(pCurrent))) {
			fChannelSendCommand(_VBUS_P pCurrent, pCmd);
			return;
		}

	}
}

#ifdef CHECK_HARDWARE_TIMEOUT
void fChannelTimeout(PChannel pChan)
{
	_VBUS_INST(pChan->pChipInstance->pVBus)
	PCommand pCmd = pChan->CurrentCmd;
	HPT_ASSERT(pCmd);
	pChan->Timer = 0;
	if (pChan->pChipInstance->ftbl.pfnResetController)
		pChan->pChipInstance->ftbl.pfnResetController(pChan);
	pCmd->Result = RETURN_BUS_RESET;
	PhysDiskDone(_VBUS_P pCmd);
	CheckPendingCall(_VBUS_P0);
}
#endif


/******************************************************************
 * Controller Send Command
 *******************************************************************/
void fChannelSendCommand(_VBUS_ARG PDevice pDevice, PCommand pCmd)
{
    PChannel          pChan = pDevice->pChannel;
    PVBus             pVBus = pChan->pChipInstance->pVBus;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
    PUCHAR            BMI = pChan->BMI;
    LBA_T			  Lba = pCmd->uCmd.Ide.Lba;
    USHORT			  nSector = pCmd->uCmd.Ide.nSectors;
    UCHAR             is48bit = 0;
    UCHAR             bIdeStatus, Count = 0;
	UCHAR old_stat;

	/* don't forget to allocate s/g table for each channel */
	HPT_ASSERT(pChan->pSgTable);

	/* call PhysDiskDone to get next command when pCmd is finished */
	HPT_ASSERT(pCmd->uScratch.disk.pfnOrgDone==0);
	pCmd->uScratch.disk.pfnOrgDone = pCmd->pfnCompletion;
	pCmd->pfnCompletion = PhysDiskDone;
	
	/*
 	 * occupy this channel & initialize the working data
 	 */
	pChan->pChWorkDev = pDevice;
	pChan->CurrentCmd = pCmd;
	pVBus->working_devs++;

	if (pChan->cf_resetting) {
    	pCmd->Result = RETURN_BUS_RESET;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return;
	}

#ifdef CHECK_HARDWARE_TIMEOUT
	pChan->Timer = AddTimerCall(_VBUS_P fChannelTimeout, pChan, HARDWARE_TIMEOUT);
#endif

    if (!pDevice->df_on_line) {
device_bad:
		KdPrint(("Device is offline"));
    	pCmd->Result = RETURN_BAD_DEVICE;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return;
	}

	if (pCmd->Result==RETURN_PENDING) pChan->bChRetry = 0;         
	pChan->cf_do_dma = 0;
	
#ifdef CHECK_CLOCK_SWITCH
	/* switch clock if needed. */
	if ((pCmd->cf_data_in || pCmd->cf_data_out) && pChan->pChipInstance->ftbl.pfnSwitchClock)
		pChan->pChipInstance->ftbl.pfnSwitchClock(pChan, pCmd->cf_data_in);
#endif


	if (!pCmd->cf_ide_passthrough) {
		/* old RAID0 has hidden lba. Remember to clear dDeHiddenLba when delete array! */
		Lba += pDevice->dDeHiddenLba;
		/* check LBA */
		if (Lba+nSector-1 > pDevice->dDeRealCapacity) {
    		pCmd->Result = RETURN_INVALID_REQUEST;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			return;
		}
	}


	/*
 	 * Select Drive 0xA0 or 0xB0 and initialize the working data
 	 */
_retry_:	
	bIdeStatus = pChan->pChipInstance->ftbl.pfnWaitOnBusy(pChan, pDevice->bDeUnitId, TRUE);

  	if(bIdeStatus & IDE_STATUS_BUSY)
  	{
		if (++pDevice->BusyCount>3)
			goto Device_Removed;
busy:
    	pCmd->Result = RETURN_DEVICE_BUSY;
    	CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
    	return;
	}
	else
		pDevice->BusyCount = 0;

	/*
 	 * check if the device was removed
 	 */
	if ( (bIdeStatus & 0x7E) == 0x7E) {
		goto Device_Removed;
	}
	else if(bIdeStatus & (IDE_STATUS_DWF|IDE_STATUS_ERROR|IDE_STATUS_DRQ)) {
		/*
		 * recover the device status if some error is found.
		 * At rare cases, some device will report DWF and DRQ status.
		 * hpt3xx: TODO: hardware polling will always has ERR bit set
		 */
		mChDisableIrq(pChan, old_stat);
		mIssueCommand(IoPort, IDE_COMMAND_RECALIBRATE);
		mChRestoreIrq(pChan, old_stat);
		mGetBaseStatus(IoPort);
		if(Count++< 10) goto _retry_;
        goto Device_Removed;
	}

	if ((bIdeStatus & IDE_STATUS_DRDY)==0) {
Device_Removed:
		KdPrint(("IDE_STATUS %x", bIdeStatus));
		pDevice->df_on_line = 0;
		pCmd->pVDevice->vf_online = 0;
		if (pCmd->pVDevice->pfnDeviceFailed) 
			CallWhenIdle(_VBUS_P (DPC_PROC)pCmd->pVDevice->pfnDeviceFailed, pCmd->pVDevice);
		fNotifyGUI(ET_DEVICE_REMOVED, Map2pVDevice(pDevice));
		goto device_bad;
	}

	/* IDE Pass through command */
	if (pCmd->cf_ide_passthrough) {
		if ((pCmd->uCmd.Passthrough.bDriveHeadReg & 0xB0)!=pDevice->bDeUnitId)
			goto invalid_ide_cmd;
		switch(pCmd->uCmd.Passthrough.bCommandReg) {
		case IDE_COMMAND_SMART:
			switch(pCmd->uCmd.Passthrough.bFeaturesReg) {
			case SMART_READ_VALUES:
			case SMART_READ_THRESHOLDS:
			case SMART_READ_LOG_SECTOR:
				if (pCmd->cf_data_in==0 || pCmd->cf_data_out)
					goto invalid_ide_cmd;
				break;
			case SMART_WRITE_LOG_SECTOR:
				if (pCmd->cf_data_in || pCmd->cf_data_out==0)
					goto invalid_ide_cmd;
				break;
			default:
				if (pCmd->cf_data_in || pCmd->cf_data_out)
					goto invalid_ide_cmd;
				break;
			}
			break;
		case IDE_COMMAND_IDENTIFY:
			if (pCmd->cf_data_in==0 || pCmd->cf_data_out)
				goto invalid_ide_cmd;
			break;
		/*for pass through extension */
		case IDE_COMMAND_READ:
			if (pCmd->cf_data_in==0 || pCmd->cf_data_out)
				goto invalid_ide_cmd;
			break;
		case IDE_COMMAND_WRITE:
			if (pCmd->cf_data_in || pCmd->cf_data_out==0)
				goto invalid_ide_cmd;
			break;
		case IDE_COMMAND_VERIFY:
			if (pCmd->cf_data_in || pCmd->cf_data_out)
				goto invalid_ide_cmd;
			break;
		case IDE_COMMAND_FLUSH_CACHE:
			break;
		default:
invalid_ide_cmd:
			pCmd->Result = RETURN_INVALID_REQUEST;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			return;
		}
		/* only allow non-sg PIO data transfer for pass through commands */
		if (pCmd->cf_data_in || pCmd->cf_data_out) {
			pCmd->uScratch.disk.WordsLeft = pCmd->uCmd.Passthrough.nSectors<<8;
			pChan->pSgTable[0].dSgAddress = (ULONG_PTR)pCmd->uCmd.Passthrough.pDataBuffer;
			pChan->pSgTable[0].wSgSize = pCmd->uCmd.Passthrough.nSectors<<9;
			pChan->pSgTable[0].wSgFlag = SG_FLAG_EOT;
			pCmd->uScratch.disk.pPIOSg = pChan->pSgTable;
		}
		/* issue command */
		mSetBlockCount(IoPort, pCmd->uCmd.Passthrough.bSectorCountReg);
		mSetBlockNumber(IoPort, pCmd->uCmd.Passthrough.bLbaLowReg);
		mSetCylinderLow(IoPort, pCmd->uCmd.Passthrough.bLbaMidReg);
		mSetCylinderHigh(IoPort, pCmd->uCmd.Passthrough.bLbaHighReg);
		mSelectUnit(IoPort, pCmd->uCmd.Passthrough.bDriveHeadReg);
		mSetFeaturePort(IoPort, pCmd->uCmd.Passthrough.bFeaturesReg);
		mIssueCommand(IoPort, pCmd->uCmd.Passthrough.bCommandReg);
		if (pCmd->cf_data_out) {
			if (fWaitForDrq(ControlPort) & IDE_STATUS_DRQ) {
				fPioInterrupt(pDevice);
				return;
			}
			else {
				pCmd->Result = RETURN_WRITE_NO_DRQ;
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			}
		}
		return;
	}

    /*
     * send ATA command
     */
    if (Lba >= 0x10000000) {
        mSetBlockCount(IoPort, (UCHAR)(nSector>>8));
        mSetBlockNumber(IoPort, (UCHAR)(Lba>>24));
        mSetCylinderLow(IoPort, 0);
        mSetCylinderHigh(IoPort,0);
        is48bit = 1;
        Lba &= 0xFFFFFF;
        Lba |= 0xE0000000;
	}
	else {
		if (pDevice->df_lba)
			Lba |= 0xE0000000;
		else
			Lba = fAsMapLbaToCHS(Lba, pDevice->wDeRealHeadXsect, pDevice->bDeRealSector);
	}
	mSetBlockCount(IoPort, (UCHAR)nSector);
	mSetBlockNumber(IoPort, (UCHAR)(Lba & 0xFF));
	mSetCylinderLow(IoPort, (UCHAR)((Lba >> 8) & 0xFF));
	mSetCylinderHigh(IoPort,(UCHAR)((Lba >> 16) & 0xFF));
	mSelectUnit(IoPort,(UCHAR)((Lba >> 24) | (pDevice->bDeUnitId)));

	if (__fWaitOnBusy(ControlPort) & IDE_STATUS_BUSY)
    	goto busy;

    /*
     * Non DMA processing
     */
	if(pDevice->df_dma==0 || pCmd->uCmd.Ide.Command==IDE_COMMAND_VERIFY) {
		if (pCmd->cf_data_in || pCmd->cf_data_out) {
pio:
			if (pCmd->pSgTable && pCmd->cf_physical_sg==0) {
				FPSCAT_GATH sg1=pChan->pSgTable, sg2=pCmd->pSgTable;
				do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
			}
			else {
				if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, pChan->pSgTable, 1)) {
					pCmd->Result = RETURN_NEED_LOGICAL_SG;
					CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
					return;
				}
			}
			pCmd->uScratch.disk.WordsLeft = pCmd->uCmd.Ide.nSectors<<8;
			pCmd->uScratch.disk.pPIOSg = pChan->pSgTable;
		}
		switch (pCmd->uCmd.Ide.Command) {
		case IDE_COMMAND_VERIFY:
    		mIssueCommand(IoPort, is48bit? IDE_COMMAND_VERIFY_EXT : IDE_COMMAND_VERIFY);
    		break;
    	case IDE_COMMAND_READ:
			pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDevice, 
				(UCHAR)MIN(pDevice->bDeModeSetting, pDevice->bDeBestPIO));
			mIssueCommand(IoPort, (is48bit)? 
				((pDevice->bDeReadCmd==IDE_COMMAND_READ)? 
					IDE_COMMAND_READ_EXT : IDE_COMMAND_READ_MULTIPLE_EXT)
				: pDevice->bDeReadCmd);
			break;
    	case IDE_COMMAND_WRITE:
			pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDevice, 
				(UCHAR)MIN(pDevice->bDeModeSetting, pDevice->bDeBestPIO));
			mIssueCommand(IoPort, (is48bit)? 
				((pDevice->bDeWriteCmd==IDE_COMMAND_WRITE)? 
					IDE_COMMAND_WRITE_EXT : IDE_COMMAND_WRITE_MULTIPLE_EXT)
				: pDevice->bDeWriteCmd);
			if (!(fWaitForDrq(ControlPort) & IDE_STATUS_DRQ)) {
				pCmd->Result = RETURN_WRITE_NO_DRQ;
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			}
			else {
				fPioInterrupt(pDevice);
			}
			break;
		default:
			pCmd->Result = RETURN_INVALID_REQUEST;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			break;
		}
    	return;
	}




	if (pCmd->pSgTable && pCmd->cf_physical_sg) {
		FPSCAT_GATH sg1=pChan->pSgTable, sg2=pCmd->pSgTable;
		do { *sg1++=*sg2; } while ((sg2++->wSgFlag & SG_FLAG_EOT)==0);
	}
	else {
		if (!pCmd->pfnBuildSgl || !pCmd->pfnBuildSgl(_VBUS_P pCmd, pChan->pSgTable, 0))
			goto pio;
	}

	OutPort(BMI, BMI_CMD_STOP);
	OutPort(BMI + BMI_STS, BMI_STS_INTR);  


#ifdef SUPPORT_DUAL_WRITE
    if(pCmd->cf_dual_write)	{
		if(is48bit)
			pCmd->cf_set_48bit = 1;
		pCmd->cf_dual_dma = 1;
		return;
	}
#endif

	if (pCmd->cf_data_in)
		mIssueCommand(IoPort, is48bit? IDE_COMMAND_READ_DMA_EXT : IDE_COMMAND_DMA_READ);
	else
		mIssueCommand(IoPort, is48bit? IDE_COMMAND_WRITE_DMA_EXT : IDE_COMMAND_DMA_WRITE);

    
	pChan->cf_do_dma = 1;

	if (IS_SUPPORTED_DEVICEID(4)) {
		if (pChan->pChipInstance->bChipType==CHIP_TYPE_HPT370A) {
			void HPT370AFixSgTable(FPSCAT_GATH pSg);
			HPT370AFixSgTable(pChan->pSgTable);
		}
	}

	if(sizeof(SCAT_GATH) > sizeof(SCAT_GATH32)) {
		/* DAC is not tested, use SAC now */
		FPSCAT_GATH32 psg1 = (FPSCAT_GATH32)pChan->pSgTable;
		FPSCAT_GATH	psg = pChan->pSgTable;
		do {
			/*HPT_ASSERT(psg->dSgAddress<=0xfffffffful); */
			psg1->dSgAddress = (ULONG) psg->dSgAddress;
			psg1->wSgSize = psg->wSgSize;
			psg1->wSgFlag = psg->wSgFlag;
			psg1++;
		}
		while ((psg++->wSgFlag & SG_FLAG_EOT)==0);
	}
	/* convert BigEndian */
	if (CPU_TO_LE32(1)!=1) {
		FPSCAT_GATH32 psg = (FPSCAT_GATH32)pChan->pSgTable;
		do {
                     	psg->dSgAddress = CPU_TO_LE32(psg->dSgAddress);
			psg->wSgSize = CPU_TO_LE16(psg->wSgSize);
			psg->wSgFlag = CPU_TO_LE16(psg->wSgFlag);
		}
		while ((psg++->wSgFlag & CPU_TO_LE16(SG_FLAG_EOT))==0);
	}

    OutDWord(pChan->BMI + BMI_DTP, pChan->SgPhysicalAddr);
		OutPort(BMI, (UCHAR)(pCmd->cf_data_in? BMI_CMD_STARTREAD : BMI_CMD_STARTWRITE));
}

/******************************************************************
 * Controller Interrupt
 ******************************************************************/
#if defined(_I960_) || defined(_WINCE_)
#define IF_ODD_PIO(x) x
#define IIF_ODD_PIO(x) if (x)
#else 
#define IF_ODD_PIO(x)
#define IIF_ODD_PIO(x) if (0)
#endif
#ifdef _MACOSX_
extern BOOLEAN fPioInterrupt(PDevice pDevice);
#else 
BOOLEAN fPioInterrupt(PDevice pDevice)
{
	PChannel pChan = pDevice->pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PCommand          pCmd = pChan->CurrentCmd;
    ULONG             wCount;
	UINT  SgWords, ThisWords;
	FPSCAT_GATH pSG = (FPSCAT_GATH)pCmd->uScratch.disk.pPIOSg;
	ADDRESS addr;

	wCount = MIN(pCmd->uScratch.disk.WordsLeft, pDevice->wWordsPerIntr);
	pCmd->uScratch.disk.WordsLeft -= wCount;

	while(wCount > 0) {
		IF_ODD_PIO (UINT onebyte);
    	if((SgWords = pSG->wSgSize) == 0) {
    		IF_ODD_PIO(onebyte = 0);
			SgWords = 0x8000;
      	}
      	else {
      		IF_ODD_PIO (onebyte = SgWords & 1);
			SgWords >>= 1;
		}

    	ThisWords = MIN(SgWords, (UINT)wCount);

		addr = os_kmap_ptr(pSG->dSgAddress);
    	if(pCmd->cf_data_out)
        	RepOUTS(IoPort, addr, ThisWords);
    	else
        	RepINS(IoPort, addr, ThisWords);
		os_kunmap_ptr(addr);

    	if((SgWords -= (USHORT)ThisWords) == 0) {
    		IIF_ODD_PIO(onebyte) {
    			WORD w = InWord(IoPort);
    			((BYTE FAR *)(ULONG_PTR)pSG->dSgAddress)[ThisWords<<1] = (BYTE)w;
    			pSG++;
    			((BYTE FAR *)(ULONG_PTR)pSG->dSgAddress)[0] = (BYTE)(w>>8);
    			pSG->dSgAddress++;
    			pSG->wSgSize--;
    			wCount -= ThisWords+1;
    		}
    		else {
        		wCount -= ThisWords;
        		pSG++;
        	}
    	}
    	else {
        	pSG->dSgAddress += (ThisWords * 2);
           	pSG->wSgSize -= (USHORT)(ThisWords * 2);
        	break;
    	}
	}
	pCmd->uScratch.disk.pPIOSg = pSG;
	
	/* return TRUE if need wait next interrupt. */
	if (pCmd->uScratch.disk.WordsLeft || pCmd->cf_data_out)
		return TRUE;

	return FALSE;
}
#endif

BOOLEAN fChannelInterrupt(_VBUS_ARG PChannel pChan)
{
    PDevice           pDevice = pChan->pChWorkDev;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PUCHAR            BMI = pChan->BMI;
    PCommand          pCmd = pChan->CurrentCmd;
    ULONG             wCount;
    UCHAR             bIdeStatus, bErrorCode;
    UCHAR	old_stat;
    
	if(pDevice == 0)
	{
		KdPrint(("fake interrupt"));
    	if(pChan->pDevice[0]) {
        	mSelectUnit(IoPort, 0xA0);
	    	mGetBaseStatus(IoPort);
	    }
    	if(pChan->pDevice[1]) {
        	mSelectUnit(IoPort, 0xB0);
			mGetBaseStatus(IoPort);
    	}
    	OutPort(BMI + BMI_STS, BMI_STS_INTR);
		return(TRUE);
	}

	if (pChan->pChipInstance->ftbl.pfnFinishInterrupt) {
		if (!pChan->pChipInstance->ftbl.pfnFinishInterrupt(pDevice))
			return FALSE;
	}
	else {
		if(pChan->cf_do_dma)
			OutPort(BMI, BMI_CMD_STOP);
	}

    /*
     * select the working device
     * clear the interrupt flags in our chip
     */
	do 
	{
		mSelectUnit(IoPort, pDevice->bDeUnitId);
		bIdeStatus = mGetBaseStatus(IoPort);
		OutPort(BMI + BMI_STS, BMI_STS_INTR);        
	} while(InPort(BMI + BMI_STS) & BMI_STS_INTR);
			
    /*
     * wait busy flag in device status off
     */
	if (bIdeStatus & IDE_STATUS_BUSY)
	{
    	for (wCount = 0; wCount < 10; wCount++)
    	{
			bIdeStatus = mGetBaseStatus(IoPort);
         	if ((bIdeStatus & IDE_STATUS_BUSY) == 0)
				goto check_drq;
			StallExec(5000);
		}
        /*
         * Let os handle this case
         */
    	/* TODO: OS_Busy_Handle(pChan, pDevice); */
    	return(TRUE);
 	}

check_drq:
    if((bIdeStatus & 0x7E) == 0x7E) /*rxl 2002/1/10  */
    {
       	pCmd->Result = RETURN_BAD_DEVICE;
		CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
		return TRUE;
	}


    /*
     * PIO operation
     */
	if (pChan->cf_do_dma==0) {
		if (bIdeStatus & IDE_STATUS_DRQ) {
			/* return if PIO read/write not done. */
			if (fPioInterrupt(pDevice)) return TRUE;
		}
		
		pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDevice, pDevice->bDeModeSetting);
	}
	else
		pChan->cf_do_dma = 0;
	
    if(bIdeStatus & IDE_STATUS_ERROR)
	{
		UCHAR Status;
		bErrorCode = mGetErrorCode(IoPort);
		KdPrint(("IDE_STATUS_ERROR 0x%x (retry %d)", bErrorCode, pChan->bChRetry));
		
		
		if (bErrorCode & 0x50) goto fail;
		
		mChDisableIrq(pChan, old_stat);
		for(wCount = 0; wCount < 10; wCount++) {
            mSelectUnit(IoPort, pDevice->bDeUnitId);
            Status = __fWaitOnBusy(pChan->BaseIoAddress2);
			if(Status & IDE_STATUS_ERROR) {
			    mIssueCommand(IoPort, IDE_COMMAND_RECALIBRATE);
				Status = __fWaitOnBusy(pChan->BaseIoAddress2);
				mGetBaseStatus(IoPort);
				OutPort(BMI + BMI_STS, BMI_STS_INTR);        
			}
			else
				break;
		}
		mChRestoreIrq(pChan, old_stat);

		
		if (pCmd->cf_ide_passthrough) {
			pDevice->wResetCount = 0;
			pCmd->Result = RETURN_IDE_ERROR;
			goto get_regs;
		}

		if(pChan->bChRetry++ > 10)
		{
			
			if (pDevice->bDeModeSetting>=8) {
				pDevice->bDeModeSetting--;
				if (pDevice->df_harddisk) fDeSetDevice(pDevice);
				fDeSelectMode(pDevice, pDevice->bDeModeSetting);
				pChan->bChRetry = 0;
			}
			else {
fail:
				pCmd->Result = RETURN_BAD_DEVICE;
				pDevice->df_on_line = 0;
				pCmd->pVDevice->vf_online = 0;
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
				if (pCmd->pVDevice->pfnDeviceFailed) 
					CallWhenIdle(_VBUS_P (DPC_PROC)pCmd->pVDevice->pfnDeviceFailed, pCmd->pVDevice);
				fNotifyGUI(ET_DEVICE_REMOVED, Map2pVDevice(pDevice));
				return TRUE;
			}
		}
		/* retry command */
    	pCmd->Result = RETURN_RETRY;
	}
	else {
		pChan->bChRetry = 0;
		pDevice->wResetCount = 0; /* clear on every command. */
		pCmd->Result = RETURN_SUCCESS;
		/* fill in register values for pass through command */
		if (pCmd->cf_ide_passthrough) {
get_regs:
			pCmd->uCmd.Passthrough.bCommandReg = bIdeStatus;
			pCmd->uCmd.Passthrough.bLbaMidReg = mGetByteLow(IoPort);
			pCmd->uCmd.Passthrough.bLbaHighReg = mGetByteHigh(IoPort);
			pCmd->uCmd.Passthrough.bDriveHeadReg = mGetUnitNumber(IoPort);
			pCmd->uCmd.Passthrough.bFeaturesReg = mGetErrorCode(IoPort);
			pCmd->uCmd.Passthrough.bSectorCountReg = mGetBlockCount(IoPort);
			pCmd->uCmd.Passthrough.bLbaLowReg = mGetBlockNumber(IoPort);
		}
    }  	

	CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);

    return(TRUE);
}

/******************************************************************
 * Chip Interrupt
 *******************************************************************/

BOOLEAN HPTLIBAPI fChipInterrupt(PChipInstance pCI)
{
    UINT     i;
    BOOLEAN  ret = 0;
	_VBUS_INST(pCI->pVBus);


    for(i = 0; i < pCI->bCpMaxChannel; i++) {
        if (InPort(pCI->pChannel[i]->BMI + BMI_STS) & BMI_STS_INTR) {
            fChannelInterrupt(_VBUS_P pCI->pChannel[i]);
            ret = TRUE;
        }
	}
    return(ret);
}

/***************************************************************************
 * Description:  Wait Device Busy Off
 ***************************************************************************/

UCHAR __fWaitOnBusy(PIDE_REGISTERS_2 BaseIoAddress)
{
	UCHAR Status;
    ULONG i;

    for (i=0; i<20000; i++)
    {
		Status = mGetStatus(BaseIoAddress);
        if ((Status & IDE_STATUS_BUSY) == 0 || Status == 0xFF)
			return Status;
        StallExec(100);
    }
    KdPrint(("WaitOnBusy(%X) timeout", BaseIoAddress));
	return(Status);
}
UCHAR fWaitOnBaseBusy(PIDE_REGISTERS_1 BaseIoAddress) 
{ 
	UCHAR Status;
	ULONG i; 
	for (i=0; i<20000; i++) { 
		Status = mGetBaseStatus(BaseIoAddress); 
		if ((Status & IDE_STATUS_BUSY)  == 0)
			return Status;
		StallExec(150); 
	}
    KdPrint(("WaitOnBaseBusy(%X) timeout", BaseIoAddress));
	return(Status); 
}

/******************************************************************
 * Wait Device DRQ on
 *******************************************************************/

UCHAR fWaitForDrq(PIDE_REGISTERS_2 BaseIoAddress)
{
    UCHAR Status;
    int  i;

    for (i=0; i<2000; i++)
    {
		Status = mGetStatus(BaseIoAddress);
        if ((Status & (IDE_STATUS_BUSY | IDE_STATUS_DRQ)) == IDE_STATUS_DRQ)
			return Status;
        StallExec(150);
    }
    KdPrint(("WaitForDrq(%X) timeout", BaseIoAddress));
    return(Status);
}
