/***************************************************************************
 * File:          device.c
 * Description:   device setting routines
 * Copyright (c)  2000-2003 HighPoint Technologies, Inc. All rights reserved
 ***************************************************************************/

#include "global.h"

/***************************************************************************
 * Function:     void fDeviceSendCommand(PDevice pDevice, void *pJob)
 * Description:  put a new job into the device queue.
 ***************************************************************************/
void HPTLIBAPI fDeviceSendCommand(_VBUS_ARG PCommand pCmd)
{
	PVDevice pVDevice = pCmd->pVDevice;
	PChannel pChan = pVDevice->u.disk.pChannel;
	PVBus pVBus = pChan->pChipInstance->pVBus;

	/* pCmd must be READ/WRITE/VERIFY. */
	if (!pCmd->cf_ide_passthrough)
		pVDevice->u.disk.HeadPosition = pCmd->uCmd.Ide.Lba + pCmd->uCmd.Ide.nSectors;


	
	if (pChan->pChipInstance->bChipType==CHIP_TYPE_HPT374 && pCmd->cf_data_out)
		pVBus->serial_mode = 1;

	if (pVBus->serial_mode) {
		if (pVBus->working_devs) {
			if (!PutCommandToQueue(&pVDevice->u.disk, pCmd)) {
        		KdPrint(("Queue full"));
	    		pCmd->Result = RETURN_DEVICE_BUSY;
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
			}
		}
 		else
 		{
			fChannelSendCommand(_VBUS_P &pVDevice->u.disk, pCmd);
 		}
	}
    else if (pChan->pChWorkDev
#ifdef CHECK_CLOCK_SWITCH
		|| (pChan->pChipInstance->ftbl.pfnClockCapable &&
			(pCmd->cf_data_in || pCmd->cf_data_out) &&
			!pChan->pChipInstance->ftbl.pfnClockCapable(pChan, pCmd->cf_data_in))
#endif
    ){
        if (!PutCommandToQueue(&pVDevice->u.disk, pCmd)) {
        	KdPrint(("Queue full"));
	    	pCmd->Result = RETURN_DEVICE_BUSY;
			CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
        }
    }
 	else
 	{
		fChannelSendCommand(_VBUS_P &pVDevice->u.disk, pCmd);
 	}
}

int IdeHardReset(PIDE_REGISTERS_1 DataPort, PIDE_REGISTERS_2 ControlPort)
{
    ULONG i;
    UCHAR dev;

	mUnitControl(ControlPort,IDE_DC_RESET_CONTROLLER|IDE_DC_DISABLE_INTERRUPTS );
	StallExec(50000L);
	mUnitControl(ControlPort,IDE_DC_REENABLE_CONTROLLER);
	StallExec(50000L);

	if (__fWaitOnBusy(ControlPort) & IDE_STATUS_BUSY)
		return FALSE;

	for (dev = 0xA0; dev<=0xB0; dev+=0x10) {
		for (i = 0; i < 100; i++) {
			mSelectUnit(DataPort, dev);
			if(mGetUnitNumber(DataPort) == dev) {
				mGetBaseStatus(DataPort);
				break;
			}
			StallExec(1000);
		}
	}
	return TRUE;
}

int fNonIoAtaCmd(PDevice pDevice, UCHAR CommandCode)
{
    PChannel pChan = pDevice->pChannel;
	PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
	UCHAR   cnt=0;

_retry_:
	mSelectUnit(IoPort, pDevice->bDeUnitId);
	if (mGetUnitNumber(IoPort) != pDevice->bDeUnitId && cnt++<100) {
		StallExec(200);
		goto _retry_;
	}
	__fWaitOnBusy(pChan->BaseIoAddress2);
	mIssueCommand(IoPort, CommandCode);
	StallExec(1000);
	__fWaitOnBusy(pChan->BaseIoAddress2);
	OutPort(pChan->BMI + BMI_STS, BMI_STS_ERROR|BMI_STS_INTR);
	return(mGetBaseStatus(IoPort));
}

UINT fGetMediaStatus(PDevice pDev)
{
	return ((fNonIoAtaCmd(pDev, IDE_COMMAND_GET_MEDIA_STATUS) << 8) |
			mGetErrorCode(pDev->pChannel->BaseIoAddress1));
}

UCHAR fSetAtaCmd(PDevice pDev, UCHAR cmd)
{
	PChannel   pChan = pDev->pChannel;
	PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
	UCHAR state, i;

	mIssueCommand(IoPort, cmd);
	StallExec(1000);
	__fWaitOnBusy(pChan->BaseIoAddress2);

    for(i=0; i<0x10; i++) {
		if (InPort(pChan->BMI + BMI_STS) & BMI_STS_INTR) {
			do {
				mSelectUnit(IoPort, pDev->bDeUnitId);
				state = mGetBaseStatus(IoPort);
				OutPort(pChan->BMI + BMI_STS, BMI_STS_ERROR|BMI_STS_INTR);        
			} while(InPort(pChan->BMI + BMI_STS) & BMI_STS_INTR);
			return state;
		}
		StallExec(1000);
	}
	return (mGetBaseStatus(IoPort));
}

int fDeSetDevice(PDevice pDev)
{
    PChannel   pChan = pDev->pChannel;
    PIDE_REGISTERS_1 IoPort = pChan->BaseIoAddress1;
	int   i=0;
	UCHAR status;

_retry_:
	mSelectUnit(IoPort, pDev->bDeUnitId);
	if (mGetUnitNumber(IoPort) != pDev->bDeUnitId) {
		if (i++<100) {
	    	StallExec(200);
        	goto _retry_;
        }
        else {
fail:
			/* set required members */
		    pDev->bDeReadCmd  = IDE_COMMAND_READ;
		    pDev->bDeWriteCmd = IDE_COMMAND_WRITE;
		    pDev->wWordsPerIntr = 256;
			return FALSE;
		}
	}

	/* set parameter for disk */
	mSelectUnit(IoPort, (UCHAR)(pDev->bDeUnitId | (pDev->bDeRealHeader-1)));
	mSetBlockCount(IoPort,  (UCHAR)pDev->bDeRealSector);
	status = fSetAtaCmd(pDev, IDE_COMMAND_SET_DRIVE_PARAMETER);
	if (status & IDE_STATUS_ERROR)
		goto fail;

    /* recalibrate */
    status = fSetAtaCmd(pDev, IDE_COMMAND_RECALIBRATE);
	if (status & IDE_STATUS_ERROR)
		goto fail;

	/* 88i8030 seems ok for r/w multiple. */
    if (pDev->wWordsPerIntr>=512 /* && pDev->df_sata==0 */) {
        /* Set to use multiple sector command */
        mSetBlockCount(IoPort,  (UCHAR)(pDev->wWordsPerIntr >> 8));
		mSelectUnit(IoPort, pDev->bDeUnitId);
        if (!(fSetAtaCmd(pDev, IDE_COMMAND_SET_MULTIPLE) & (IDE_STATUS_BUSY | IDE_STATUS_ERROR))) {
            pDev->bDeReadCmd  = IDE_COMMAND_READ_MULTIPLE;
            pDev->bDeWriteCmd = IDE_COMMAND_WRITE_MULTIPLE;
            pDev->df_multiple = 1;
            return TRUE;
         }
    }
    pDev->bDeReadCmd  = IDE_COMMAND_READ;
    pDev->bDeWriteCmd = IDE_COMMAND_WRITE;
    pDev->wWordsPerIntr = 256;
    return TRUE;
}

/******************************************************************
 * Select Device Mode
 *******************************************************************/
void HPTLIBAPI fDeSelectMode(PDevice pDev, UCHAR NewMode)
{
    PChannel   pChan = pDev->pChannel;
    PIDE_REGISTERS_1 IoPort = pChan->BaseIoAddress1;
	UCHAR      Feature;
	int        i=0;

_retry_:
	mSelectUnit(IoPort, pDev->bDeUnitId);
	if ((mGetUnitNumber(IoPort) != pDev->bDeUnitId) && i++<100) {
		StallExec(200);
		goto _retry_;
	}

	if (IS_SATA_DEVICE(pDev)) {
		
		if (NewMode>4 && NewMode<8) NewMode = 4;
		
		if (NewMode>8 && NewMode<12) NewMode = 8;
		
		
		
	}

    /* Set Feature */
    mSetFeaturePort(IoPort, 3);
	if(NewMode < 5) {
		pDev->df_dma = pDev->df_ultra = 0;
		Feature = (UCHAR)(NewMode | FT_USE_PIO);
	} else if(NewMode < 8) {
        pDev->df_dma = 1;
		Feature = (UCHAR)((NewMode - 5)| FT_USE_MWDMA);
	} else {
        pDev->df_dma = pDev->df_ultra = 1;
		Feature = (UCHAR)((NewMode - 8) | FT_USE_ULTRA);
		if (NewMode>14 && (pChan->pChipInstance->ChipFlags & CHIP_LIMIT_U7)) Feature--;
    }

    mSetBlockCount(IoPort, Feature);
	fSetAtaCmd(pDev, IDE_COMMAND_SET_FEATURES);

	pDev->bDeModeSetting = NewMode;
	pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDev, NewMode);

}

/******************************************************************
 * IO ATA Command
 *******************************************************************/

int HPTLIBAPI fDeReadWrite(PDevice pDev, ULONG Lba, UCHAR Cmd, void *tmpBuffer)
{
	PChannel   pChan = pDev->pChannel;
	PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
	PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
	UCHAR      statusByte;
    UINT       i, retry=0;
    UCHAR old_stat;
    int is48bit = (Lba & 0xF0000000)!=0;

	UCHAR old_mode = pDev->bDeModeSetting;

	
	mChDisableIrq(pChan, old_stat);
	fDeSelectMode(pDev, 0);

	i=0;
_retry_:
	mSelectUnit(IoPort, pDev->bDeUnitId);
    if (mGetUnitNumber(IoPort) != pDev->bDeUnitId && i++<100) {
        StallExec(200);
        goto _retry_;
    }
    if (i>=100) goto out;
	__fWaitOnBusy(ControlPort);

    if (is48bit) {
        mSetBlockCount(IoPort, 0);
        mSetBlockNumber(IoPort, (UCHAR)(Lba>>24));
        mSetCylinderLow(IoPort, 0);
        mSetCylinderHigh(IoPort,0);
	}
	else if(pDev->df_lba)
		Lba |= 0xE0000000; /* will it cause error when retry? */
	else
		Lba = fAsMapLbaToCHS(Lba, pDev->wDeRealHeadXsect, pDev->bDeRealSector);

	mSetBlockCount(IoPort, 1);
	mSetBlockNumber(IoPort, (UCHAR)(Lba & 0xFF));
	mSetCylinderLow(IoPort, (UCHAR)((Lba >> 8) & 0xFF));
	mSetCylinderHigh(IoPort,(UCHAR)((Lba >> 16) & 0xFF));
	mSelectUnit(IoPort, is48bit? (0xE0|pDev->bDeUnitId) : (UCHAR)((Lba>>24) | pDev->bDeUnitId));

	__fWaitOnBusy(ControlPort);

	if (is48bit)
		mIssueCommand(IoPort, Cmd==IDE_COMMAND_READ?IDE_COMMAND_READ_EXT:IDE_COMMAND_WRITE_EXT);
	else
		mIssueCommand(IoPort, Cmd);

	for(i = 0; i < 5; i++)	{
		statusByte = __fWaitOnBusy(ControlPort);
		if((statusByte & (IDE_STATUS_BUSY | IDE_STATUS_ERROR)) == 0)
			goto check_drq;
	}
out:
    if (retry++<4) {
		statusByte= mGetErrorCode(IoPort);
		mIssueCommand(IoPort, IDE_COMMAND_RECALIBRATE);
		mGetBaseStatus(IoPort);
		StallExec(10000);
        goto _retry_;
    }
    fDeSelectMode(pDev, old_mode);
	mChRestoreIrq(pChan, old_stat);
	return(FALSE);

check_drq:
	if((statusByte & IDE_STATUS_DRQ) == 0) {
		statusByte = fWaitForDrq(ControlPort);
		if((statusByte & IDE_STATUS_DRQ) == 0)	{
			mGetBaseStatus(IoPort); /*Clear interrupt */
			goto out;
		}
	}
	mGetBaseStatus(IoPort); /*Clear interrupt */

	if(Cmd == IDE_COMMAND_READ)
		RepINS(IoPort, (ADDRESS)tmpBuffer, 256);
	else {
		RepOUTS(IoPort, (ADDRESS)tmpBuffer, 256);
	}

	fDeSelectMode(pDev, old_mode);
	mChRestoreIrq(pChan, old_stat);
	return(TRUE);
}

#ifdef SUPPORT_HPT601			
void BeginAccess601(PIDE_REGISTERS_1 IoPort) 
{
	int i;
	for(i=0; i<5; i++) InPort(&IoPort->CylinderLow);
	for(i=0; i<6; i++) InPort(&IoPort->CylinderHigh);
	for(i=0; i<5; i++) InPort(&IoPort->CylinderLow);
}
void EndAccess601(PIDE_REGISTERS_1 IoPort)
{
	OutPort(&IoPort->BlockNumber, 0x80);
}
#endif

extern int autorebuild;
extern void hpt_check_autorebuild(_VBUS_ARG PVDevice pVDev);
void fRescanAllDevice(_VBUS_ARG int quick_probe)
{
	PChipInstance	pChip;
	PChannel		pChan;
	PVDevice 		pVDevice;
	PDevice			pDevice;
	int				j, iDev;
	UINT			NewFlag = 0;
	UCHAR 			old_stat;
	UINT            vbus;
	PVBus _vbus_p = gVBus;

	for(vbus=0; vbus<_vbus_p->nInstances; vbus++) {
		pChip = _vbus_p->pChipInstance[vbus];

		for(j = 0; j < pChip->bCpMaxChannel; j++)
		{
			pChan = pChip->pChannel[j];
			if(pChan->pChWorkDev != NULL) continue;

			mChDisableIrq(pChan, old_stat);
			for(iDev = 0 ; iDev < 2; iDev++)
			{
#ifdef IS_SATA_CHANNEL
				if (iDev && IS_SATA_CHANNEL(pChan)) continue;
#endif
				pDevice = pChan->pDevice[iDev];
				if (pDevice && pDevice->df_on_line)
				{
					/* Check if device is still working. */
					PIDE_REGISTERS_1 IoPort = pChan->BaseIoAddress1;
					UCHAR StatusByte, cnt = 0;
					
					if (quick_probe) continue;

_retry_:
					/* 
					 * send a command to drive to detect if it works.
					 * This is required for S-ATA chip.
					 */
					mSelectUnit(IoPort, pDevice->bDeUnitId);
					if (!pDevice->df_atapi)
						mIssueCommand(IoPort, IDE_COMMAND_FLUSH_CACHE);
					StatusByte=pChip->ftbl.pfnWaitOnBusy(pChan, pDevice->bDeUnitId, FALSE);
					mGetBaseStatus(IoPort);
					
					if ((StatusByte & 0x7E) == 0x7E)
					{
						goto device_removed;
					}
					else if (StatusByte & IDE_STATUS_DWF) {
						/*
						 * gmm 2001-3-18
						 * Some disks will set IDE_STATUS_DWF for a while
						 * when the other disk on same channel get removed
						 */
						if (pDevice->df_atapi)
							goto device_removed;
						StatusByte= mGetErrorCode(IoPort);
						mIssueCommand(IoPort, IDE_COMMAND_RECALIBRATE);
						mGetBaseStatus(IoPort);
						StallExec(10000);
						if((cnt++) < 10)
							goto _retry_;
						StatusByte = 0x7e; /* cause it to be removed */
					}
					else if(StatusByte & IDE_STATUS_ERROR)
					{
						StatusByte = mGetErrorCode(IoPort);
						mIssueCommand(IoPort, IDE_COMMAND_RECALIBRATE);
						mGetBaseStatus(IoPort);
						if((cnt++) < 10)
							goto _retry_;
						StatusByte = 0x7e; /* cause it to be removed */
					}

					if((StatusByte & 0x7e)== 0x7e || ((StatusByte & (IDE_STATUS_DRDY|IDE_STATUS_BUSY))==0))
					{
device_removed:
						pVDevice = Map2pVDevice(pDevice);
						pDevice->df_on_line = 0;
						pVDevice->vf_online = 0;
						if (pVDevice->pfnDeviceFailed) {
							_VBUS_INST(pChip->pVBus)
							CallWhenIdle(_VBUS_P (DPC_PROC)pVDevice->pfnDeviceFailed, pVDevice);
						}
						fNotifyGUI(ET_DEVICE_REMOVED, pVDevice);
					}
				}
				else /* !pDevice || !pDevice->df_on_line */
				{
					if(pDevice == NULL)
					{
						if((pDevice = OsGetDeviceTable(pChan,pDevice)) != 0)
						{
							pVDevice = Map2pVDevice(pDevice);
							pChan->pDevice[iDev] = pDevice;
            				pDevice->pChannel = pChan;
            				pVDevice->pVBus = pChan->pChipInstance->pVBus;
			                pDevice->bDeUnitId = (BYTE)((iDev << 4) | 0xA0);
			                pDevice->wWordsPerIntr = 256;
			                pDevice->DeviceID = (UCHAR)((pChip->bChipID + j) * 2 + iDev);
            				NewFlag = 1;
						}
						else continue;
					}

					if (fDeScanDevice(pDevice, quick_probe))
            		{
#ifdef SUPPORT_ARRAY
            			DECLARE_BUFFER(PArrayDescript, pArrayDesc);
#endif
            			pVDevice = Map2pVDevice(pDevice);

						if (pDevice->df_harddisk)
							fDeSetDevice(pDevice);
						/* 
						 * give the chip a chance to adjust clock;
						 * driver shall be idle at this time.
						 */
						if (pChip->ftbl.pfnInit2) pChip->ftbl.pfnInit2(pChip);
						pVDevice->vf_online = 1;
						pVDevice->VDeviceType = VD_SINGLE_DISK;

						if (pDevice->df_harddisk) {
#ifdef SUPPORT_ARRAY
							if(pVDevice->pParent) {
								int iMember;
								for(iMember = 0; iMember < 	pVDevice->pParent->u.array.bArnMember; iMember++)
									if((PVDevice)pVDevice->pParent->u.array.pMember[iMember] == pVDevice)
										pVDevice->pParent->u.array.pMember[iMember] = NULL;
								pVDevice->pParent = NULL;
							}
#ifdef ARRAY_V2_ONLY
							pVDevice->vf_format_v2 = 1;
#else 
							pVDevice->vf_format_v2 = 0;
#endif
							pVDevice->VDeviceCapacity = pDevice->dDeRealCapacity-SAVE_FOR_RAID_INFO;
							/* Check Array Block after above fields set */
							fDeReadWrite(pDevice, pDevice->dDeRealCapacity-SAVE_FOR_RAID_INFO, IDE_COMMAND_READ, pArrayDesc);
							if((pArrayDesc->Signature!=HPT_ARRAY_V3)
								|| ((pArrayDesc->Signature==HPT_ARRAY_V3) && (pArrayDesc->dArStamp == 0)))
								fCheckBootable(pVDevice);
#else 
							pVDevice->VDeviceCapacity = pDevice->dDeRealCapacity;
#endif
						}
						else
	                        pVDevice->VDeviceType = VD_ATAPI;

						pVDevice->pfnSendCommand = pfnSendCommand[pVDevice->VDeviceType];
						pVDevice->pfnDeviceFailed = pfnDeviceFailed[pVDevice->VDeviceType];
						fNotifyGUI(ET_DEVICE_PLUGGED, pVDevice);

						if (autorebuild) {
							_VBUS_INST(pChip->pVBus)
							CallWhenIdle(_VBUS_P (DPC_PROC)hpt_check_autorebuild, pVDevice);
						}
						else 
						{
							RegisterVDevice(pVDevice);
						}
					}
					else if(NewFlag)
					{
						pChan->pDevice[iDev] = 0;
                		OsReturnTable(pDevice);
                	}
                	NewFlag = 0;
				}
			}
			mChRestoreIrq(pChan, old_stat);
		}
	}
}
