#include <linux/config.h>
#include <linux/version.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)) && defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/sched.h>
#include <linux/module.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <linux/blk.h>
#else 
#include <linux/moduleparam.h>
#endif

#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <asm/types.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/pci.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/genhd.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>
#include <linux/timer.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) /* 0x20300 */
	#include <asm/spinlock.h>
#else 
	#include <linux/spinlock.h>
#endif
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,2,16) 
#include <linux/kcomp.h>
#endif
#include <linux/kernel.h>
#include <linux/smp_lock.h>
#include <asm/dma.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <linux/reboot.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#include <linux/nmi.h>
#endif

#if defined(__MODULE_KERNEL_i686) && defined(__MODULE_KERNEL_athlon) && (__MODULE_KERNEL_i686==__MODULE_KERNEL_athlon)
#error "__MODULE_KERNEL_i686 and __MODULE_KERNEL_athlon is mistakely defined"
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include "scsi.h"
#include <scsi/scsi_ioctl.h>
#else 
#include <linux/blkdev.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi.h>
typedef struct scsi_host_template Scsi_Host_Template;
typedef struct scsi_device Scsi_Device;
typedef struct scsi_cmnd Scsi_Cmnd;
typedef struct scsi_request Scsi_Request;
typedef struct scsi_pointer Scsi_Pointer;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,4)
#include "hosts.h"
#else 
#include <scsi/scsi_host.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define set_mb(var, value) do { xchg(&var, value); } while (0)

#define __set_current_state(state_value)			\
	do { current->state = (state_value); } while (0)
#ifdef CONFIG_SMP
#define set_current_state(state_value)		\
	set_mb(current->state, (state_value))
#else 
#define set_current_state(state_value)		\
	__set_current_state(state_value)
#endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include "sd.h"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define scsi_to_pci_dma_dir(scsi_dir) ((int)(scsi_dir))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#define hpt_verify_area(type, addr, size) (!access_ok((type), (addr), (size)))
#else 
#define hpt_verify_area(type, addr, size) (verify_area((type), (addr), (size)))
#endif

#ifdef SUPPORT_ARRAY
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,4)
#define __BDEV_RAW
#else 
#define __BDEV_RAW , BDEV_RAW
#endif
static struct block_device * get_bdev(struct Scsi_Host *host, int id)
{
	static int major[] = { SCSI_DISK0_MAJOR, SCSI_DISK1_MAJOR, SCSI_DISK2_MAJOR, SCSI_DISK3_MAJOR, 
				SCSI_DISK4_MAJOR, SCSI_DISK5_MAJOR, SCSI_DISK6_MAJOR, SCSI_DISK7_MAJOR, 
				SCSI_DISK8_MAJOR, SCSI_DISK9_MAJOR, SCSI_DISK10_MAJOR, SCSI_DISK11_MAJOR, 
				SCSI_DISK12_MAJOR, SCSI_DISK13_MAJOR, SCSI_DISK14_MAJOR, SCSI_DISK15_MAJOR, 0 };
	Scsi_Device *SDptr;
	shost_for_each_device(SDptr, host) {
		if (SDptr->id==id) {
			scsi_device_put(SDptr);
			break;
		}
	}
	if (SDptr) {
		int i, minor;
		for (i=0; major[i]; i++) {
			for (minor=0; minor<=240; minor++) {
				struct block_device *bdev = bdget(MKDEV(major[i], minor));
				if (bdev && blkdev_get(bdev, FMODE_READ, 0 __BDEV_RAW)==0) {
					if (bdev->bd_disk && bdev->bd_disk->driverfs_dev==&SDptr->sdev_gendev) {
						return bdev;
					}
					blkdev_put(bdev __BDEV_RAW);
				}
			}
		}
	}
	return 0;
}
static int sd_inuse(struct Scsi_Host *host, int id)
{
	struct block_device * bdev = get_bdev(host, id);
	int inuse = 0;
	if (bdev) {
		if (bdev->bd_openers>1) inuse=1;
		blkdev_put(bdev __BDEV_RAW);
	}
	return inuse;
}

#else 

static int sd_inuse(struct Scsi_Host *host, int id)
{
	Scsi_Device *SDptr = host->host_queue;
	while (SDptr) {
		if ((SDptr->id == id) && SDptr->access_count)
			return 1;
		SDptr = SDptr->next;
	}
	return 0;
}
#endif
#endif

#if LINUX_VERSION_CODE  < KERNEL_VERSION(2,4,4) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) 
#define scsi_set_pci_device(pshost, pcidev)	do { ;} while (0)
#endif

#if defined(RR1520)
#define DRIVER_NAME "RocketRAID 1520 SATA Controller driver"
#else 
#define DRIVER_NAME "HPT37x2 RAID Controller driver"
#endif
#define PROC_DIR_NAME "hpt37x2"

#include "entry.c"
#include "hptproc.c"
#include "ioctl.c"

#include "channel.c"
#include "device.c"
#include "hpt3xx.c"
#include "idevice.c"
#include "ichip.c"  
           
#include "vdevice.c"
#include "init.c"

#ifdef RR1520
void Hpt3xxSataReset(PChannel pChan)
{
	PUCHAR BMI = pChan->pChipInstance->pPciIoMap;
	int i;
	/* 
	 * How to locate the RST_ pin for the boards?
	 * R1540: PRST_(ch1), SRST_(ch2), TRST_(x), FRST_(ch3,4)
	 * RR154x: PRST_(ch1,3), SRST_(ch2), TRST_(buzzer), FRST_(ch4)
	 * RR1520/R15xx/R1110/RR1680: one-to-one
	 * HPT371+SATA: SRST_ links to ATAIOSEL, don't use!
	 */
	UCHAR oldmisc;
	UCHAR rst_bit = pChan==pChan->pChipInstance->pChannel[0]? 0x40: 0x80;
	
	oldmisc = InPort(BMI+0x79);
	KdPrint(("hpt3xx: enable RST_ signal (%X->%X)", oldmisc, oldmisc|rst_bit));
	OutPort(BMI+0x79, (UCHAR)(oldmisc|rst_bit));
	StallExec(10000);
	OutPort(BMI+0x79, oldmisc);
	StallExec(60000);StallExec(40000);/* fool driver verifier: StallExec(100000); */
	
	/* IDE status may become ready after a long time. Found on RR1544. */
	if (pChan->pDevice[0]) i=100; else i=20;
	for (; i>0; i--) {
		if (mGetBaseStatus(pChan->BaseIoAddress1)==0x50) break;
		StallExec(50000);
	}
}
#endif
