/***************************************************************************
 * File:          Hpt3xx.c
 * Description:   HPT3XX IC handlers
 * Copyright (c)  2000-2003 HighPoint Technologies, Inc. All rights reserved
 ***************************************************************************/

#include "global.h"


#include "hpt3xx.h"

void Hpt3xxModeSetting(PChannel pChan, PDevice pDevice, UCHAR NewMode)
{
	OutDWord((PULONG)(pChan->BMI + ((pDevice->bDeUnitId & 0x10)>>2) + 0x60),
        (ULONG)(pChan->pChipInstance->Setting[NewMode]));
}

UCHAR Hpt3xxControlIrq(PChipInstance pChip, UCHAR enable)
{
	UCHAR old_stat = InPort(pChip->pPciIoMap+0x7A) & 0x10;
	OutPort(pChip->pPciIoMap+0x7A, (UCHAR)(enable? 0 : 0x10));
    return (old_stat & 0x10)? 0 : 1;
}

/***************************************************************************
 * Function:     void HPT3xxCableDetect(PDevice pDevice)
 * Description:  detect if the cable is 80 ping cable
 * Dependence:   device.h
 * Source file:  device.c
 * Argument:     PDevice pDevice - The device that
 * Retures:      None
 ***************************************************************************/

void fHPT3xxCableDetect(PDevice pDevice)
{
    PChannel pChan = pDevice->pChannel;
	UCHAR	 ucOldSetting;
	int iChan = pChan - pChan->pChipInstance->pChannel[0];
	PUCHAR BaseBMI = pChan->pChipInstance->pPciIoMap;
	PUCHAR misc;

	if (pChan->cf_80pin_cable) {
		/* enable MA15,MA16 as input pins to detect whether is 80pin cable */
		if (pChan->pChipInstance->pci_func) {
			misc = BaseBMI + (iChan? 0x77 : 0x73);
			ucOldSetting = InPort(misc);
			OutPort(misc, (UCHAR)(ucOldSetting|0x80));
		}
		else {
			misc = BaseBMI + 0x7B;
			ucOldSetting = InPort(misc);
			OutPort(misc, (UCHAR)(ucOldSetting & 0xFE));
		}
	
		StallExec(10); /* wait awhile for debouncing. */
		if (InPort(BaseBMI + 0x7A) & (iChan? 1 : 2)) pChan->cf_80pin_cable = 0;
		OutPort(misc, ucOldSetting);
	}
}

/* IDE engine reset. */
void fHptReset370IdeEngine(PChannel pChan, UCHAR UnitId)
{
	if(pChan->pChipInstance->bChipType > CHIP_TYPE_HPT370B) /* HPT372 and later */
	{
        PUCHAR tmpBMI = pChan->pChipInstance->pPciIoMap + 
			(((UINT_PTR)pChan->BMI & 0xF)? 0x74 : 0x70);

        OutPort(tmpBMI+3, 0x80);
        OutPort(tmpBMI, 0x37);
		OutPort(tmpBMI+3, 0);
	}
	else {
		PULONG  SettingPort;
		ULONG   OldSettings, tmp;
		PUCHAR  tmpBMI = pChan->BMI + (((UINT_PTR)pChan->BMI & 0xF)? 0x6C : 0x70);
		PUCHAR tmpBaseBMI = pChan->pChipInstance->pPciIoMap+0x79;
		UCHAR  oldmisc = InPort(tmpBaseBMI);

		OutPort(tmpBMI+3, 0x80);
		/* tri-state the IDE bus */
		OutPort(tmpBaseBMI, (UCHAR)(((UINT_PTR)tmpBMI & 0xF)? 0x80 : 0x40));
		/* reset state machine */
		OutPort(tmpBMI, 0x37);
		StallExec(2);
		
		SettingPort= (PULONG)(pChan->BMI+((UnitId &0x10) >>2)+ 0x60);
		OldSettings= InDWord(SettingPort);
		tmp= OldSettings & 0xfffffff;
		OutDWord(SettingPort, (tmp|0x80000000));
		OutWord(pChan->BaseIoAddress1, 0x0);
		StallExec(10);
		OutPort(tmpBMI, 0x37);
		StallExec(2);
		OutDWord(SettingPort, OldSettings);
		/* set IDE bus to normal state */
		OutPort(tmpBaseBMI, oldmisc);
		OutPort(tmpBMI+3, 0);
	}
}

#ifdef CHECK_CLOCK_SWITCH


int fHPT370ClockCapable(PChannel pChan, int reading)
{
	PChipInstance pChip = pChan->pChipInstance;
	int thisChan = pChan - pChip->pChannel[0];
	PChannel pOther = pChip->pChannel[thisChan^1];
	UCHAR need_dpll = (reading==0 && thisChan);
	return (pOther->CurrentCmd==0 || need_dpll==pChip->using_dpll);
}

/* set chip clock for reading or writing */
void fHPT370SwitchClock(PChannel pChan, int reading)
{
    PChipInstance  pChip = pChan->pChipInstance;
    PUCHAR BaseBMI = pChip->pPciIoMap;

    PUCHAR PrimaryMiscCtrlRegister = BaseBMI + 0x70;
    PUCHAR SecondaryMiscCtrlRegister = BaseBMI + 0x74;
    UCHAR oldmisc;

	UCHAR need_dpll = (reading==0 && (pChan-pChip->pChannel[0]));
	if (pChip->using_dpll==need_dpll) return;

	pChip->using_dpll = need_dpll;

    OutPort(PrimaryMiscCtrlRegister+3, 0x80); /* tri-state the primary channel */
    OutPort(SecondaryMiscCtrlRegister+3, 0x80); /* tri-state the secondary channel */

    OutPort(BaseBMI+0x7B, (UCHAR)(need_dpll? 0x21: 0x23)); /* switching the clock */

	oldmisc = InPort(BaseBMI+0x79);
    OutPort(BaseBMI+0x79, (UCHAR)(oldmisc|0xC0)); /* reset two channels begin */

    OutPort(PrimaryMiscCtrlRegister, 0x37); /* reset primary channel state machine */
    OutPort(SecondaryMiscCtrlRegister, 0x37); /* reset secordary channel state machine */

    OutPort(BaseBMI+0x79, oldmisc); /* reset two channels finished             */

    OutPort(PrimaryMiscCtrlRegister+3, 0x00); /* normal-state the primary channel */
    OutPort(SecondaryMiscCtrlRegister+3, 0x00); /* normal-state the secondary channel */
}


int fHPT3x2NClockCapable(PChannel pChan, int reading)
{
	PChipInstance pChip = pChan->pChipInstance;
	int thisChan = pChan - pChip->pChannel[0];
	PChannel pOther = pChip->pChannel[thisChan^1];
	UCHAR need_dpll;
	PDevice pDev = pChan->pDevice[0];
	if (pDev && pDev->bDeModeSetting>14)
		need_dpll = reading==0 || thisChan;
	else
		need_dpll = reading==0;
	
	return (pOther->CurrentCmd==0 || need_dpll==pChip->using_dpll);
}

/* set chip clock for reading or writing */
void fHPT3x2NSwitchClock(PChannel pChan, int reading)
{
    PChipInstance  pChip = pChan->pChipInstance;
    PUCHAR BaseBMI = pChip->pPciIoMap;
	int thisChan = pChan - pChip->pChannel[0];

    PUCHAR PrimaryMiscCtrlRegister = BaseBMI + 0x70;
    PUCHAR SecondaryMiscCtrlRegister = BaseBMI + 0x74;

	UCHAR need_dpll, oldmisc;
	PDevice pDev = pChan->pDevice[0];
	if (pDev && pDev->bDeModeSetting>14)
		need_dpll = reading==0 || thisChan;
	else
		need_dpll = reading==0;
	if (pChip->using_dpll==need_dpll) return;

	pChip->using_dpll = need_dpll;

    OutPort(PrimaryMiscCtrlRegister+3, 0x80); /* tri-state the primary channel */
    OutPort(SecondaryMiscCtrlRegister+3, 0x80); /* tri-state the secondary channel */

    OutPort(BaseBMI+0x7B, (UCHAR)(need_dpll? 0x21: 0x23)); /* switching the clock */

	oldmisc = InPort(BaseBMI+0x79);
    OutPort(BaseBMI+0x79, (UCHAR)(oldmisc|0xC0)); /* reset two channels begin */

    OutPort(PrimaryMiscCtrlRegister, 0x37); /* reset primary channel state machine */
    OutPort(SecondaryMiscCtrlRegister, 0x37); /* reset secordary channel state machine */

    OutPort(BaseBMI+0x79, oldmisc); /* reset two channels finished             */

    OutPort(PrimaryMiscCtrlRegister+3, 0x00); /* normal-state the primary channel */
    OutPort(SecondaryMiscCtrlRegister+3, 0x00); /* normal-state the secondary channel */
}

#endif

BOOLEAN fHpt3xxFinishIntr(PDevice pDev)
{
	PChannel pChan = pDev->pChannel;
	if(pChan->cf_do_dma) {
        /*
         * check if the DMA has done.
		 * if FIFO count in misc 3 register NOT equ 0, it's a fake interrupt.
         */
    	if(InWord(pChan->MiscControlAddr+2) & 0x1FF)
			return FALSE;
		
        
    	OutPort(pChan->BMI, BMI_CMD_STOP);
    	fHptReset370IdeEngine(pChan, pDev->bDeUnitId);
	}
    return TRUE;
}

UCHAR fHpt3xxWaitOnBusy(PChannel pChan, UCHAR unit, int select)
{
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
	UINT i;
	UCHAR status;

	if (select) {
		for (i = 0; i < 100; i++) {
			mSelectUnit(IoPort, unit);
      		if((UCHAR)mGetUnitNumber(IoPort) == unit) {
        		break;
			}
    		StallExec(100);
		}
		
		if (i>=100)	{
			
			Hpt3xxSataReset(pChan);
			mSelectUnit(IoPort, unit);
			__fWaitOnBusy(ControlPort);
			mSetBlockCount(IoPort, 0x7F);
		}
	}

	status = __fWaitOnBusy(ControlPort);
#ifdef HPT3XX_SATA_RESET
	/* found problem on RM1210: status busy after drive was removed, reset can recover */
	if ((status & IDE_STATUS_BUSY)) {
		KdPrint(("status is 0x%x, reset", status));
		Hpt3xxSataReset(pChan);
		mSelectUnit(IoPort, unit);
		status = __fWaitOnBusy(ControlPort);
	}
#endif
	return status;
}

void fHpt3xxResetController(PChannel pChan)
{
	int i;
	PDevice pDev;
	PVDevice pVDev;
    PIDE_REGISTERS_1 IoPort = pChan->BaseIoAddress1;
    PIDE_REGISTERS_2 ControlPort = pChan->BaseIoAddress2;
    UCHAR old_stat;
    int reset_failed;

    mChDisableIrq(pChan, old_stat);
	if (pChan->pChipInstance->ftbl.pfnResetEngine)
		pChan->pChipInstance->ftbl.pfnResetEngine(pChan, 0xA0);
	Hpt3xxSataReset(pChan);
	reset_failed = !IdeHardReset(IoPort, ControlPort);

    for(i = 0; i < 2; i++) {
        if((pDev = pChan->pDevice[i]) == 0)	continue;
        if (pDev->df_on_line) {
			/* reduce mode for DMA timeout, disable the device for PIO timeout */
			if (pChan->pChWorkDev==pDev) {
				if (pChan->cf_do_dma) {
					if (++pDev->wResetCount>1) {
						if (pDev->bDeModeSetting) pDev->bDeModeSetting--;
						pDev->wResetCount = 0;
					}
				}
				else if (++pDev->wResetCount>6) {
					pVDev = Map2pVDevice(pDev);
					pDev->df_on_line = 0;
					pVDev->vf_online = 0;
					if (pVDev->pfnDeviceFailed) {
						_VBUS_INST(pVDev->pVBus)
						CallWhenIdle(_VBUS_P (DPC_PROC)pVDev->pfnDeviceFailed, pVDev);
					}
					fNotifyGUI(ET_DEVICE_REMOVED, pVDev);
					continue;
				}
			}
			if (pDev->df_harddisk) fDeSetDevice(pDev);
			fDeSelectMode(pDev, pDev->bDeModeSetting);
		}
	}
	mChRestoreIrq(pChan, old_stat);
}

void HPT370AFixSgTable(FPSCAT_GATH pSg)
{
	
	if (pSg[0].wSgSize==0 && (pSg[0].wSgFlag & SG_FLAG_EOT)) {
		pSg[0].wSgSize = 0x8000;
		pSg[0].wSgFlag = 0;
		pSg[1].dSgAddress = pSg[0].dSgAddress + 0x8000;
		pSg[1].wSgSize = 0x8000;
		pSg[1].wSgFlag = SG_FLAG_EOT;
	}
}


#ifdef SUPPORT_DUAL_WRITE
void Hpt3xxStartDualDma(PCommand pSubCmd, PCommand pDualCmd)
{
	PChannel          pChan1 = pSubCmd->pVDevice->u.disk.pChannel;
	PIDE_REGISTERS_1  IoPort1 = pChan1->BaseIoAddress1;
    PUCHAR            BMI1 = pChan1->BMI;
	PChannel          pChan2 = pDualCmd->pVDevice->u.disk.pChannel;
	PIDE_REGISTERS_1  IoPort2 = pChan2->BaseIoAddress1;
    PUCHAR            BMI2 = pChan2->BMI;

    OutPort((((UINT)pChan1->BMI & ~0xF)+ 0x9C), 1);
	mIssueCommand(IoPort1, pSubCmd->cf_set_48bit? IDE_COMMAND_WRITE_DMA_EXT : IDE_COMMAND_DMA_WRITE);
	mIssueCommand(IoPort2, pDualCmd->cf_set_48bit? IDE_COMMAND_WRITE_DMA_EXT : IDE_COMMAND_DMA_WRITE);
    pChan1->cf_do_dma = 1;
    pChan2->cf_do_dma = 1;
    OutDWord(BMI1 + BMI_DTP , pChan1->SgPhysicalAddr);
    OutDWord(BMI2 + BMI_DTP , pChan1->SgPhysicalAddr);
    OutPort(BMI1, BMI_CMD_STARTWRITE);
    OutPort(BMI2, BMI_CMD_STARTWRITE);
}
void Hpt3xxStartSingleDma(PCommand pCmd)
{
	PChannel          pChan = pCmd->pVDevice->u.disk.pChannel;
	PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PUCHAR            BMI = pChan->BMI;

	pCmd->cf_dual_dma = 0;
	pCmd->cf_dual_write = 0;

	mIssueCommand(IoPort, pCmd->cf_set_48bit? IDE_COMMAND_WRITE_DMA_EXT : IDE_COMMAND_DMA_WRITE);
    pChan->cf_do_dma = 1;

    OutDWord(BMI + BMI_DTP , pChan->SgPhysicalAddr);
    OutPort(BMI, BMI_CMD_STARTWRITE);
}
#endif


