/*
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * Copyright (C) 2000-2003 HighPoint Technologies, Inc.
 * proc.c  /proc file system support
 */
#include "global.h"
#include "hptintf.h"
#include "oslinux.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/*
 * proc entry
 */
struct proc_dir_entry proc_scsi_hpt3xx = { 
	0, 							/* inode number - ignore. filled by proc_register_dynamic */
	sizeof(PROC_DIR_NAME)-1, 	/* length of name */
	PROC_DIR_NAME,              /* file name */
	S_IFDIR|S_IRUGO|S_IXUGO,	/* */
	2							/* number of links */
};
#endif

/*************************************************************************
 * hpt_set_asc_info
 * Usage:
 *   echo "hpt rebuild start" > /proc/scsi/hpt37x2/0
 *   echo "hpt rebuild stop" > /proc/scsi/hpt37x2/0
 *   echo "hpt verify start x" > /proc/scsi/hpt37x2/0
 *   echo "hpt verify stop x" > /proc/scsi/hpt37x2/0
 *   echo "hpt dbglevel x" > /proc/scsi/hpt37x2/0
 *        x can be 0, 1, 2
 *************************************************************************/
int hpt_rescan_all(void);
int hpt_add_disk_to_array(DEVICEID idArray, DEVICEID idDisk);

int hpt_set_asc_info(char *buffer,int length)
{
	int orig_length = length+4;

#ifdef SUPPORT_ARRAY	
    unsigned long flags;
	PVBus pVBus;
	PVDevice pArray;
	PVDevice pSubArray, pVDev;
    UINT iarray, ichan, idev, i, j;
    
	if (length>=8 && strncmp(buffer, "rebuild ", 8)==0) {
		buffer+=8;
		length-=8;
		if (length>=5 && strncmp(buffer, "start", 5)==0) {
			/* start rebuild */
			spin_lock_irqsave(&driver_lock, flags);
			FOR_EACH_ARRAY_ON_ALL_VBUS(pVBus, pArray, i) {
				if (pArray->u.array.rf_need_rebuild && !pArray->u.array.rf_rebuilding)
                    hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pArray, 
						(UCHAR)((pArray->u.array.CriticalMembers || pArray->VDeviceType == VD_RAID_1)? DUPLICATE : REBUILD_PARITY));
			}
			spin_unlock_irqrestore(&driver_lock, flags);
			return orig_length;
		}
		else if (length>=4 && strncmp(buffer, "stop", 4)==0) {
			/* stop rebuild */
			spin_lock_irqsave(&driver_lock, flags);
			FOR_EACH_ARRAY_ON_ALL_VBUS(pVBus, pArray, i) {
				if (pArray->u.array.rf_rebuilding)
				    pArray->u.array.rf_abort_rebuild = 1;
			}
			spin_unlock_irqrestore(&driver_lock, flags);
			return orig_length;
		}
		else if (length>=5 && buffer[1]==','&& buffer[3]==','&& buffer[0]>='0')
		{
		    /* hpt rebuild x,x,x */
            iarray = buffer[0]-'1';
	        ichan = buffer[2]-'1';
	        idev = buffer[4]-'0';

			pArray = gVBus[0].pVDevice[iarray];
	        if (!pArray || (pArray->vf_online == 0)) return -EINVAL;

			if(idev > 1) return -EINVAL;

			for(i = 0; i < gVBus[0].nInstances; i++)
		        for(j = 0; j < MAX_CHANNEL_PER_CHIP; j++)
			        if((i*MAX_CHANNEL_PER_CHIP + j) == ichan)
			            goto rebuild;
	        return -EINVAL;

rebuild:
	        pVDev = Map2pVDevice(gVBus[0].pChipInstance[i]->pChannel[j]->pDevice[idev]);
	        if(!pVDev || !pVDev->u.disk.df_on_line || pVDev->pParent) return -EINVAL;

	        /* Not allow to use a mounted disk ??? test*/
			for(i = 0; i < MAX_DEVICES; i++)
			    if(pVDev == gVBus[0].pVDevice[i])
			    {
			        spin_lock_irqsave(&driver_lock, flags);
			        if (sd_inuse(hpt_vhost, i))
					   {
						   hpt_printk(("Can not delete a mounted device.\n"));
						   return -1;
					   }
					spin_unlock_irqrestore(&driver_lock, flags);
			     }
			switch(pArray->VDeviceType) {
		    case VD_RAID_1:
			case VD_RAID_5:
			{
			    pSubArray = pArray;
loop:
                if(hpt_add_disk_to_array((ULONG)(ULONG_PTR)pSubArray, (ULONG)(ULONG_PTR)pVDev) == -1) return -EINVAL;
	            break;
			}
			case VD_RAID_0:
				for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++) 
					if(pArray->u.array.pMember[i] && mIsArray(pArray->u.array.pMember[i]) &&
					   (pArray->u.array.pMember[i]->u.array.rf_broken == 1))
					{
						  pSubArray = pArray->u.array.pMember[i];
						  goto loop;
					}
			default:
				return -EINVAL;
			}
			return orig_length;
		}
	}
	else if (length>=7 && strncmp(buffer, "verify ", 7)==0) {
		buffer+=7;
		length-=7;
        if (length>=6 && strncmp(buffer, "start ", 6)==0) {
            buffer+=6;
		    length-=6;
            if (length>=1 && *buffer>='0') {
				iarray = *buffer-'1';
				pArray = gVBus[0].pVDevice[iarray];
				if (!pArray || (pArray->vf_online == 0)) return -EINVAL;
				if(pArray->VDeviceType != VD_RAID_1 && pArray->VDeviceType != VD_RAID_5)
					return -EINVAL;
				if (!(pArray->u.array.rf_need_rebuild ||
					pArray->u.array.rf_rebuilding ||
					pArray->u.array.rf_verifying ||
					pArray->u.array.rf_initializing))
				{
					spin_lock_irqsave(&driver_lock, flags);
					pArray->u.array.RebuildSectors = 0;
					hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pArray, VERIFY_R);
					spin_unlock_irqrestore(&driver_lock, flags);
				}
                return orig_length;
			}
		}
		else if (length>=5 && strncmp(buffer, "stop ", 5)==0) {
            buffer+=5;
		    length-=5;
            if (length>=1 && *buffer>='0') {
				iarray = *buffer-'1';
				pArray = gVBus[0].pVDevice[iarray];
				if (!pArray || (pArray->vf_online == 0)) return -EINVAL;
				if(pArray->u.array.rf_verifying) {
					spin_lock_irqsave(&driver_lock, flags);
				    pArray->u.array.rf_abort_rebuild = 1;
					spin_unlock_irqrestore(&driver_lock, flags);
				}
			    return orig_length;
			}
		}
    }
	else 
#endif
	if (length>=7 && strncmp(buffer, "rescan ", 7)==0) {
		buffer+=7;
		length-=7;
		if (length>=3 && strncmp(buffer, "all", 3)==0) {
			hpt_rescan_all();
			return orig_length;
		} else if (length>=5 && buffer[1]==',' && buffer[3]==',') {
            /*do nothing now*/
			return orig_length;
		}
	}
#if DBG
	else if (length>=9 && strncmp(buffer, "dbglevel ", 9)==0) {
	    buffer+=9;
		length-=9;
		if (length>=1 && *buffer>='0' && *buffer<='3') {
			hpt_dbg_level = *buffer-'0';
			return orig_length;
		}
	}
#endif
	return -EINVAL;
}

int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						);

int hpt_set_info(char *buffer,int length)
{
	int retval;
#ifdef SUPPORT_IOCTL
	PUCHAR ke_area;
	int err;
	DWORD dwRet;
	DWORD dwIoControlCode;
	void *lpInBuffer;
	DWORD nInBufferSize;
	void *lpOutBuffer;
	DWORD nOutBufferSize;
	void *lpBytesReturned;
#endif

	if (length >= 4) {
		if (strncmp(buffer,"hpt ",4) == 0) {
			return (retval = hpt_set_asc_info(buffer+4, length-4)) >= 0? retval : -EINVAL;
		}

#ifdef SUPPORT_IOCTL
		if (*(DWORD *)buffer == HPT_IOCTL_MAGIC) {
			dwIoControlCode = ((PHPT_IOCTL_PARAM)buffer)->dwIoControlCode;
			lpInBuffer = ((PHPT_IOCTL_PARAM)buffer)->lpInBuffer;
			nInBufferSize = ((PHPT_IOCTL_PARAM)buffer)->nInBufferSize;
			lpOutBuffer = ((PHPT_IOCTL_PARAM)buffer)->lpOutBuffer;
			nOutBufferSize = ((PHPT_IOCTL_PARAM)buffer)->nOutBufferSize;
			lpBytesReturned = ((PHPT_IOCTL_PARAM)buffer)->lpBytesReturned;
		}
		else if (*(DWORD *)buffer == HPT_IOCTL_MAGIC32) {
			dwIoControlCode = ((PHPT_IOCTL_PARAM32)buffer)->dwIoControlCode;
			lpInBuffer = (void*)(ULONG_PTR)((PHPT_IOCTL_PARAM32)buffer)->lpInBuffer;
			nInBufferSize = ((PHPT_IOCTL_PARAM32)buffer)->nInBufferSize;
			lpOutBuffer = (void*)(ULONG_PTR)((PHPT_IOCTL_PARAM32)buffer)->lpOutBuffer;
			nOutBufferSize = ((PHPT_IOCTL_PARAM32)buffer)->nOutBufferSize;
			lpBytesReturned = (void*)(ULONG_PTR)((PHPT_IOCTL_PARAM32)buffer)->lpBytesReturned;
		}
		else return -EINVAL;

        KdPrintE(("ioctl=%d in=%x len=%d out=%x len=%d\n", 
			dwIoControlCode, lpInBuffer, nInBufferSize, lpOutBuffer, nOutBufferSize));

			/* 
        	 * verify user buffer.
        	 */
    	if (nInBufferSize) {
			if (hpt_verify_area(VERIFY_READ, (void *)(ULONG_PTR)lpInBuffer, nInBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
			}
		}
		if (nOutBufferSize) {
			if (hpt_verify_area(VERIFY_WRITE, (void *)(ULONG_PTR)lpOutBuffer, nOutBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
			}
		}
		if (lpBytesReturned) {
			if (hpt_verify_area(VERIFY_WRITE, (void *)(ULONG_PTR)lpBytesReturned, sizeof(DWORD))) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
			}
		}
		
		/*
         * map buffer to kernel.
         */
    	if (nInBufferSize+nOutBufferSize > 0x10000) {
        		KdPrintE(("User buffer too large\n"));
        		return -EINVAL;
        }
    	ke_area = kmalloc(nInBufferSize+nOutBufferSize, GFP_ATOMIC);
		if (ke_area == NULL) {
			KdPrintE(("Couldn't allocate kernel mem.\n"));
			return -EINVAL;
		}
		if (nInBufferSize) {
			if (copy_from_user(ke_area, (void *)(ULONG_PTR)lpInBuffer, nInBufferSize)) {
				kfree(ke_area);
				KdPrintE(("copy_from_user error\n"));
				return -EINVAL;
			}
		}
		
			/*
			 * call kernel handler.
			 */
		err = Kernel_DeviceIoControl(dwIoControlCode, ke_area, nInBufferSize,
				ke_area + nInBufferSize, nOutBufferSize, &dwRet);    
		if (err==0) {
			if (nOutBufferSize)
				if (copy_to_user((void *)(ULONG_PTR)lpOutBuffer, 
						ke_area + nInBufferSize, 
						nOutBufferSize)) {
					kfree(ke_area);
					KdPrintE(("copy_to_user 1 error\n"));
					return -EINVAL;
				}
			if (lpBytesReturned)
		    	if (copy_to_user((void*)(ULONG_PTR)lpBytesReturned, &dwRet, sizeof(DWORD))) {
					kfree(ke_area);
					KdPrintE(("copy_to_user 2 error\n"));
					return -EINVAL;
				}
			
			kfree(ke_area);
			return length;
		}
		else {
			KdPrintW(("Kernel_ioctl(): return %d\n", err));
		}
		
		kfree(ke_area);
        return -EINVAL;
#endif
    }
    
    return -EINVAL;
}

typedef struct _hpt_GET_INFO {
	char *buffer;
	int buflength;
	int bufoffset;
	int buffillen;
	int filpos;	
} HPT_GET_INFO;

void hpt_copy_mem_info(HPT_GET_INFO *pInfo, char *data, int datalen)
{
	if (pInfo->filpos < pInfo->bufoffset) {	/* Current offset before buffer offset */
		if (pInfo->filpos + datalen <= pInfo->bufoffset) {
			pInfo->filpos += datalen; 		/* Discard if completely before buffer */
			return;
		} else {				/* Partial copy, set to begin */
			data += (pInfo->bufoffset - pInfo->filpos);
			datalen  -= (pInfo->bufoffset - pInfo->filpos);
			pInfo->filpos = pInfo->bufoffset;
		}
	}
	
	pInfo->filpos += datalen;		/* Update current offset */
	if (pInfo->buffillen == pInfo->buflength) /* Buffer full, discard */
		return;
	
	if (pInfo->buflength - pInfo->buffillen < datalen)  /* Overflows buffer ? */
	datalen = pInfo->buflength - pInfo->buffillen;
	
	memcpy(pInfo->buffer + pInfo->buffillen, data, datalen);
	pInfo->buffillen += datalen;
}

#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

static void get_disk_name(char *name, PDevice pDev)
{
	int i;
	for (i = 0; i < 10; i++) 
		((WORD*)name)[i] = shortswap(pDev->IdentifyData.ModelNumber[i]);
	name[20] = '\0';
}

int hpt_copy_info(HPT_GET_INFO *pinfo, char *fmt, ...) 
{
	va_list args;
	char buf[128];
	int len;
  
	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	hpt_copy_mem_info(pinfo, buf, len);
	return len;
}

static void hpt_copy_disk_info(HPT_GET_INFO *pinfo, PVDevice pVDev)
{
	UINT iBus, iChip, iChan, iDev;
	char name[32], arrayname[16];
	get_disk_name(name, &pVDev->u.disk);
#ifdef SUPPORT_ARRAY
	if(pVDev->pParent) 
		memcpy(arrayname, pVDev->pParent->u.array.ArrayName, MAX_ARRAY_NAME);
	else
#endif
	    arrayname[0]=0;
	
	for(iBus = 0; iBus < MAX_VBUS; iBus++) 			
		for(iChip = 0; iChip < gVBus[iBus].nInstances; iChip++)
			for(iChan = 0; iChan < MAX_CHANNEL_PER_CHIP; iChan++)
	        {
				for(iDev = 0; iDev < MAX_DEVICE_PER_CHANNEL; iDev++)
				    if(&pVDev->u.disk == gVBus[iBus].pChipInstance[iChip]->pChannel[iChan]->pDevice[iDev])
					{
						 hpt_copy_info(pinfo, "%d Channel %d %s %s  %5dMB  %s %s\n",
							  (iChip+1), (iChan+1), 
							  pVDev->u.disk.df_sata? "      " : iDev?"Slave ":"Master", 
							  name, pVDev->VDeviceCapacity>>11, 
							  ((!pVDev->u.disk.df_on_line)? "Disabled" : 
							  ((pVDev->VDeviceType != VD_SPARE)?"Normal  ":"Spare   ")), arrayname);
						 return;
					}
				 
				 
			}
}

#ifdef SUPPORT_ARRAY
static void hpt_copy_array_info(HPT_GET_INFO *pinfo, int nld, PVDevice pArray, char *sd)
{
	int i;
	char *sType=0, *sStatus=0;
	char buf[32];
    PVDevice pTmpArray;

	switch (pArray->VDeviceType) {
	case VD_RAID_0:
        sType = "RAID 0     ";
		break;
	case VD_RAID_1:
		for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++) 
		  if(pArray->u.array.pMember[i])
		  {
			  if(mIsArray(pArray->u.array.pMember[i]))
				 sType = "RAID 0/1   ";
			  else
				 sType = "RAID 1     ";
			  break;
		  }
		break;
	case VD_JBOD:
		sType = "JBOD       ";
		break;
	default:
		sType = "N/A        ";
		break;
	}
	
	if (pArray->vf_online == 0)
		sStatus = "Disabled";
	else if (pArray->u.array.rf_broken)
		sStatus = "Critical";
	for (i = 0; (UCHAR)i < pArray->u.array.bArnMember; i++)
	{
	    if (!sStatus) 
		{
#ifndef ARRAY_V2_ONLY
		    if(mIsArray(pArray->u.array.pMember[i])) {
                  pTmpArray = pArray->u.array.pMember[i]; 
	   		}
		    else
#endif
		    pTmpArray = pArray;
		      if (pTmpArray->u.array.rf_rebuilding) {
				sprintf(buf, "Rebuilding %d%%", (pTmpArray->u.array.RebuildSectors>>11)*100/((pTmpArray->VDeviceCapacity/(pTmpArray->u.array.bArnMember-1))>>11));
				sStatus = buf;
			}
			else if (pTmpArray->u.array.rf_verifying) {
				sprintf(buf, "Verifying %d%%", (pTmpArray->u.array.RebuildSectors>>11)*100/((pTmpArray->VDeviceCapacity/(pTmpArray->u.array.bArnMember-1))>>11));
				sStatus = buf;
			}
			else if (pTmpArray->u.array.rf_need_rebuild)
				sStatus = "Critical";
			else if (pTmpArray->u.array.rf_broken)
				sStatus = "Critical";
			if(pTmpArray == pArray) goto out;
		}
		else
			goto out;
	}
out:	
	if (!sStatus) sStatus = "Normal";
	hpt_copy_info(pinfo, "%2d  %3s   %11s  %-20s  %5dMB  %s\n", nld, sd, sType, pArray->u.array.ArrayName, pArray->VDeviceCapacity>>11, sStatus);
}
#endif

static void get_sd_name(char *name, struct Scsi_Host *host, int id)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	Scsi_Device *SDptr;
	int first_devno = 0;
	name[0] = 0;
	SDptr = host->host_queue;
	if (SDptr && SDptr->device_queue) 
		first_devno = MINOR(SDptr->device_queue->request.rq_dev)/16;
	while (SDptr) {
	   if (SDptr->id == id) {
		   if(SDptr->device_queue) {
			   int devno = (SDptr->device_queue->request.rq_dev)?(MINOR(SDptr->device_queue->request.rq_dev)/16):(first_devno+id);
			   sprintf(name, "/dev/sd%c", 'a' + devno);
		   }
		   break;
	   }
	   SDptr = SDptr->next;
	}
#else 
	struct block_device *bdev = get_bdev(host, id);
	name[0] = 0;
	if (bdev) {
		sprintf(name, "/dev/%s", bdev->bd_disk->disk_name);
		blkdev_put(bdev __BDEV_RAW);
	}
#endif
}

int hpt_get_info(char *buffer,char **start,off_t offset,int length)
{
	HPT_GET_INFO info;
	UINT iBus,iChip,j,i;
	PDevice pDev;
	PVDevice pVDev;
	static char sd[32];
 
	info.buffer     = buffer;
	info.buflength  = length;
	info.bufoffset  = offset;
	info.filpos     = 0;
	info.buffillen  = 0;

	if (start) *start = buffer;

	hpt_copy_info(&info, "%s Version %s\n\n", DRIVER_NAME, DRIVER_VERSION);

	hpt_copy_info(&info, "Physical device list\n");

	hpt_copy_info(&info, "Controller/Bus/ID  Model                Capacity  Status   Array\n");
	hpt_copy_info(&info, "-------------------------------------------------------------------\n");

	for(iBus = 0; iBus < MAX_VBUS; iBus++) 			
		for(iChip = 0; iChip < gVBus[iBus].nInstances; iChip++)
			for(j = 0; j < MAX_CHANNEL_PER_CHIP; j++)
				for(i = 0; i < MAX_DEVICE_PER_CHANNEL; i++)
	            {
					pDev=gVBus[iBus].pChipInstance[iChip]->pChannel[j]->pDevice[i];
					if(pDev)
					   hpt_copy_disk_info(&info, Map2pVDevice(pDev));
				}

	hpt_copy_info(&info, "\nLogical device list\n");
	hpt_copy_info(&info, "No. sd         Type         Name                 Capacity  Status\n");
	hpt_copy_info(&info, "-------------------------------------------------------------------\n");
	
	j=1;
	for(iBus = 0; iBus < MAX_VBUS; iBus++) 
		for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++){
            pVDev = gVBus[iBus].pVDevice[i];
			if(pVDev){
				ZeroMemory(sd,32);
				j=i+1;
				get_sd_name(sd, hpt_vhost, i);
#ifdef SUPPORT_ARRAY
				if(mIsArray(pVDev)){
	is_array:
					hpt_copy_array_info(&info, j, pVDev, sd);
				}
				else
#endif
				{
					char name[32];
#ifdef SUPPORT_ARRAY
					/* it may be add to an array after driver loaded, check it */
					if (pVDev->pParent)
						/* in this case, pVDev can only be a RAID 1 source disk. */
						if (pVDev->pParent->VDeviceType==VD_RAID_1 && pVDev==pVDev->pParent->u.array.pMember[0]) 
							goto is_array;
#endif
					get_disk_name(name, &pVDev->u.disk);
					hpt_copy_info(&info, 
						"%2d  %3s   %s  %s  %5dMB  %s\n",
						j, sd, "Single disk", name, pVDev->VDeviceCapacity>>11, 
						/* gmm 2001-6-19: Check if pVDev has been added to an array. */
						((pVDev->pParent) ? "Unavailable" : "Normal"));
				 }
	         }
	     }
	

	
	return info.buffillen;
}

/*******************************************************************************
 * hpt3xx_ProcInfo
 * Parameters:
 *  *buffer: I/O buffer
 *  **start: if inout == FALSE pointer into buffer where user read should start
 *  offset: current offset into a /proc/scsi/hpt37x2/[0...] file
 *  length: length of buffer
 *  hostno: Scsi_Host host_no
 *  inout: TRUE - user is writing; FALSE - user is reading
 * Returns:
 *   Length of data
 *****/
int hpt3xx_ProcInfo(char *buffer,char **start,off_t offset,int length,   
                   int hostno,int inout)
{
	KdPrintI(("hpt3xx_ProcInfo() length %d ha %d offs %d inout %d\n",
		length,hostno,(int)offset,inout));

	if (inout)
		return(hpt_set_info(buffer,length));
	else
		return(hpt_get_info(buffer,start,offset,length));
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
int hpt3xx_ProcInfo26(struct Scsi_Host *host, char *buffer, char **start, 
			off_t offset, int length, int inout)
{
	return hpt3xx_ProcInfo(buffer, start, offset, length, host->host_no, inout);
}
#endif
