/***************************************************************************
 * File:          ichip.c
 * Description:   scan for chips
 * Copyright (c)  2000-2003 HighPoint Technologies, Inc. All rights reserved
 ***************************************************************************/

#include "global.h"

static UINT ChipChannels = 0; 
int pci_clock_mhz=0;
extern int is_hiber_driver;

#include "hpt3xx.h"





CHIP_CONFIG_INFO ChipInfoTable[]= {
	{ 0x41103, 3, CHIP_TYPE_HPT370,  5, &hpt370_proc_table },
	{ 0x41103, 4, CHIP_TYPE_HPT370A, 5, &hpt370_proc_table },
	{ 0x41103, 5, CHIP_TYPE_HPT372,  5, &hpt3xx_proc_table },
	{ 0x41103, 6, CHIP_TYPE_HPT372N, 5, &hpt3x2n_proc_table },
	{ 0x51103, 1, CHIP_TYPE_HPT372A, 5, &hpt3xx_proc_table },
	{ 0x51103, 2, CHIP_TYPE_HPT372N, 5, &hpt3x2n_proc_table },
	{ 0x61103, 1, CHIP_TYPE_HPT302,  5, &hpt3xx_proc_table },
	{ 0x61103, 2, CHIP_TYPE_HPT302N, 5, &hpt3x2n_proc_table },
	{ 0x71103, 1, CHIP_TYPE_HPT371,  5, &hpt3xx_proc_table },
	{ 0x71103, 2, CHIP_TYPE_HPT371N, 5, &hpt3x2n_proc_table },
	{ 0x81103, 7, CHIP_TYPE_HPT374,  5, &hpt3xx_proc_table },
};

int nChipInfo = sizeof(ChipInfoTable)/sizeof(CHIP_CONFIG_INFO);

PCHIP_CONFIG_INFO GetChipConfigInfo(ULONG id, UCHAR rev)
{
	int i;
	for (i=0; i<nChipInfo; i++) {
		if (ChipInfoTable[i].DeviceId==id && 
			(ChipInfoTable[i].Revision==rev || ChipInfoTable[i].Revision==0xFF))
			return &ChipInfoTable[i];
	}
	return 0; /* default unknown */
}

#define DumpChipInfo(pChip)

/*===================================================================
 * Scan Chip
 *===================================================================*/


static UCHAR scan_bus_num = 0;
static UCHAR scan_dev_num = 0;
static UCHAR scan_fun_num = 0;
static UCHAR first_on_slot = 1;
UINT GetNextPciInfo(PChipInstance pChip)
{
    ULONG     ChipVIdDId;
	PCHIP_CONFIG_INFO pChipCfg;
    UINT      FoundChip = 0;


    while(scan_bus_num < MAX_PCI_BUS_NUMBER) {
		do {
			while (scan_fun_num < MAX_PCI_FUNC_NUMBER) {
				
				if(FoundChip) goto out;
				
                ChipVIdDId = pcicfg_read_dword(scan_bus_num, scan_dev_num, scan_fun_num, 0);

				
				if ((USHORT)ChipVIdDId==0x1103 && !IS_SUPPORTED_DEVICEID(ChipVIdDId>>16))
					break; 

				pChipCfg = GetChipConfigInfo(ChipVIdDId, 
					pcicfg_read_byte(scan_bus_num, scan_dev_num, scan_fun_num, REG_RID));

				if (pChipCfg){
					pChip->bChipType = pChipCfg->ChipType;
					pChip->ftbl = *pChipCfg->ftbl;
					pChip->pci_bus = scan_bus_num;
					pChip->pci_dev = scan_dev_num;
					pChip->pci_func = scan_fun_num;
					pChip->first_on_slot = first_on_slot;
                    pChip->bChipIntrNum = pcicfg_read_byte(scan_bus_num, scan_dev_num, scan_fun_num, 0x3C);
					DumpChipInfo(pChip);
                    fRegisterChip(pChip);
                    FoundChip = 1;
					first_on_slot = 0;
                }
				/* check multi-function support */
				if (scan_fun_num==0 &&
					(pcicfg_read_byte(scan_bus_num, scan_dev_num, 0, 0xE) & 0x80)==0)
					break;
                scan_fun_num++;
            }

            scan_fun_num = 0;
			first_on_slot = 1;
        } while(scan_dev_num++ != MAX_PCI_DEVICE_NUMBER - 1);

        scan_dev_num = 0;
        scan_bus_num++;
    }

out:
    return (FoundChip);
}

#include "ihpt3xx.c"





