/***************************************************************************
 * File:          iDevice.c
 * Description:   scan for drives and initialize them
 * Copyright (c)  2000-2003 HighPoint Technologies, Inc. All rights reserved
 ***************************************************************************/

#include "global.h"

PBadModeList check_bad_disk(char *ModelNumber);

int fDeIssueIdentify(PDevice pDev, UCHAR Cmd, void *tmpBuffer)
{
    PChannel   pChan = pDev->pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    PIDE_REGISTERS_2  ControlPort = pChan->BaseIoAddress2;
    int      i, retry=0;

	/* use PIO 4 timing for CD-ROMs. HDDs shall be OK too, but keep old code now. */
	pChan->pChipInstance->ftbl.pfnModeSetting(pChan, pDev, 
		(UCHAR)(Cmd==IDE_COMMAND_ATAPI_IDENTIFY? 4 : 0));

_retry_:
    mSelectUnit(IoPort, pDev->bDeUnitId);
    if(__fWaitOnBusy(ControlPort) & IDE_STATUS_BUSY) {
		IdeHardReset(IoPort, ControlPort);
	}
	
	i=0;
	do {
		mSelectUnit(IoPort, pDev->bDeUnitId);
		StallExec(200);
    }
	while ((mGetUnitNumber(IoPort) != pDev->bDeUnitId) && i++<100);

    mIssueCommand(IoPort, Cmd);
	
    for(i = 0; i < 5; i++)
        if(!(__fWaitOnBusy(ControlPort) & (IDE_STATUS_ERROR |IDE_STATUS_BUSY)))
            break;

	if (i < 5 && (fWaitForDrq(ControlPort) & IDE_STATUS_DRQ)) {
		__fWaitOnBusy(ControlPort);
		mGetBaseStatus(IoPort);
		OutPort(pChan->BMI + BMI_STS, BMI_STS_ERROR|BMI_STS_INTR);
		RepINS(IoPort, (ADDRESS)tmpBuffer, 256);
		
		
		return(TRUE);
	}
	
	if (++retry < 2) goto _retry_;
	mGetBaseStatus(IoPort);
	return(FALSE);
}

/********************************************************************
 * Scan one device
 ********************************************************************/

int fDeScanDevice(PDevice pDev, int quick_probe)
{
	DECLARE_BUFFER(PIDENTIFY_DATA, pIdentify);
    PChannel          pChan = pDev->pChannel;
    PIDE_REGISTERS_1  IoPort = pChan->BaseIoAddress1;
    int               j, retry;
    UCHAR             stat;
    

#ifdef SUPPORT_HPT601
	mSelectUnit(IoPort, pDev->bDeUnitId);
	__fWaitOnBusy(pChan->BaseIoAddress2);
	BeginAccess601(IoPort);
	mSetBlockCount(IoPort, 0);
	if (InWord(&IoPort->Data) == 0x5a3e) {
		pDev->df_with_601 = 1;
		mSetBlockCount(IoPort, 0xA);/* power on fan by default */
		OutWord(&IoPort->Data, 0);
	}
	EndAccess601(IoPort);
#endif

	/* check if channel disabled? */
	if ((InPort(pChan->MiscControlAddr) & 4)==0)
		return 0;
		
	if (quick_probe) {
		mSelectUnit(IoPort, pDev->bDeUnitId);
		stat = mGetBaseStatus(IoPort);
		if (stat & IDE_STATUS_BUSY)
			return 0;
		goto check_port;
	}

	retry = 0;
	for(;;) {
		stat = pChan->pChipInstance->ftbl.pfnWaitOnBusy(pChan, pDev->bDeUnitId, 1);
		if (stat==0x7f) return 0;
		if((stat & IDE_STATUS_BUSY) == 0)
			goto check_port;
		if (++retry>3) return 0;

	}

check_port:
    mSetBlockNumber(IoPort, 0x55);
    mSetBlockCount(IoPort, 0);
    if (mGetBlockNumber(IoPort) != 0x55) {
        mSelectUnit(IoPort,(UCHAR)(pDev->bDeUnitId ^ 0x10));
        return 0;
    }
    mSetBlockNumber(IoPort, 0xAA);
    if(mGetBlockNumber(IoPort) != 0xAA)
        return 0;

    if(mGetByteLow(IoPort) == 0x14 && mGetByteHigh(IoPort) == 0xEB)
        goto is_cdrom;

    for(j = 0; j != 0xFFFF; j++) {
        stat = mGetBaseStatus(IoPort);
        if(stat & IDE_STATUS_DWF)
             break;
        if((stat & IDE_STATUS_BUSY) == 0) {
             if((stat & (IDE_STATUS_DSC|IDE_STATUS_DRDY)) == (IDE_STATUS_DSC|IDE_STATUS_DRDY))
                 goto chk_cd_again;
             break;
        }
        StallExec(5000);
    }

    if((mGetBaseStatus(IoPort) & 0xAE) != 0)
        return 0;

/*===================================================================
 * Read Identifytify data for a device
 *===================================================================*/

chk_cd_again:
    if(mGetByteLow(IoPort) == 0x14 && mGetByteHigh(IoPort) == 0xEB) {
is_cdrom:
		return 0;

    } else if(fDeIssueIdentify(pDev, IDE_COMMAND_IDENTIFY, pIdentify) == FALSE) { 

        if((mGetBaseStatus(IoPort) & ~1) == 0x50 ||
            (mGetByteLow(IoPort) == 0x14 && mGetByteHigh(IoPort) == 0xEB))
            goto is_cdrom;
        else
            return 0;
	}

	/* 2006-4-11 check power up in standby feature */
	if (((WORD *)pIdentify)[0] & CPU_TO_LE16(0x4)) {
		if (((WORD *)pIdentify)[2]==CPU_TO_LE16(0x37C8)) {

			mSetFeaturePort(IoPort, 7);
			mIssueCommand(IoPort, IDE_COMMAND_SET_FEATURES);
			for (j=0; j<20000; j++) { 
				StallExec(1000);
				stat = mGetBaseStatus(IoPort); 
				if ((stat & IDE_STATUS_BUSY)==0) {
					if ((stat & IDE_STATUS_ERROR)==0) {
						((WORD *)pIdentify)[0] = 0;
						goto chk_cd_again;
					}
					else
						break;
				}
			}
			return 0;
		}
	}
	
	/*
	 * set the df_sata flag base on different products. 
	 * P-ATA only products shall define CHECK_SATA_DEVICE(x) to null.
	 */
	CHECK_SATA_DEVICE(pDev);

	fSetDeviceTable(pDev, pIdentify);

	
	if (
		IS_SATA_DEVICE(pDev)) {
		
		/* Seagate S-ATA drives, "ST...AS" */
		if (pIdentify->ModelNumber[0]==LE16_TO_CPU(0x5354) /*'ST'*/) {
			char *p = (char *)&pIdentify->ModelNumber;
			if (*(WORD*)(p+8)==LE16_TO_CPU(0x4153)/*'AS'*/ || (p[8]=='A' && p[11]=='S')) {
				pDev->bMaxShowMode = 15;
				if (pDev->bDeUsable_Mode>13)
					pDev->bDeUsable_Mode = 13;
			}
		}
		else if ((LE16_TO_CPU(pIdentify->SataCapability) & 3)==2) {
			pDev->bMaxShowMode = 15;
			if (pIdentify->ModelNumber[0]==LE16_TO_CPU(0x5744)/*WD*/ && pDev->bDeUsable_Mode>13)
				pDev->bDeUsable_Mode = 13;
		}
	}


	if (pChan->bMaxMode && pDev->bDeUsable_Mode>pChan->bMaxMode)
		pDev->bMaxShowMode = pDev->bDeUsable_Mode = pChan->bMaxMode;

	OsSetDeviceTable(pDev, pIdentify);

	fDeSelectMode(pDev, pDev->bDeUsable_Mode);
	return 1;
}

/********************************************************************
 * Scan and Initialize all devices on the chip
 ********************************************************************/

int fDeScanAllDevices(PChipInstance pChip)
{
    UINT     i, j;
    PChannel pChan;
    PDevice  pDevice;
    int      total = 0;

    for(i = 0; i < (UINT)pChip->bCpMaxChannel; i++) 
    { 
		pChan = pChip->pChannel[i];

        for(j = 0; j < MAX_DEVICE_PER_CHANNEL; j++) 
        {
#ifdef IS_SATA_CHANNEL
			if (j && IS_SATA_CHANNEL(pChan)) break;
#endif
			if((pDevice = OsGetDeviceTable(pChan,pDevice)) != 0) 
			{ 
				pChan->pDevice[j] = pDevice;
                pDevice->pChannel = pChan;

                pDevice->bDeUnitId = (BYTE)((j << 4) | 0xA0);
                pDevice->wWordsPerIntr = 256;
                pDevice->DeviceID = (UCHAR)((pChip->bChipID + i) * 2 + j);

                if(fDeScanDevice(pDevice, 0)) 
                {				
						fDeSetDevice(pDevice);
					IdeRegisterDevice(pDevice);
                    total++;
					StallExec(500);					
                } 
                else 
                {
                    pChan->pDevice[j] = 0;
                    OsReturnTable(pDevice);
                }
            }
        } 
    }
    return(total);
}

#ifndef SUPPORT_ARRAY
void HPTLIBAPI IdeRegisterDevice(PDevice pDev)
{
	DECLARE_BUFFER(PUCHAR, sector);
	PVDevice pVDev = Map2pVDevice(pDev);
	pVDev->VDeviceType = pDev->df_atapi? VD_ATAPI : 
						 pDev->df_removable_drive? VD_REMOVABLE : VD_SINGLE_DISK;
	pVDev->pParent = 0;
	pVDev->vf_online = 1;
	pVDev->pVBus = pDev->pChannel->pChipInstance->pVBus;
	pVDev->VDeviceCapacity = pDev->dDeRealCapacity;
	pVDev->pfnSendCommand = pfnSendCommand[pVDev->VDeviceType];
	pVDev->pfnDeviceFailed = pfnDeviceFailed[pVDev->VDeviceType];
	if (!pDev->df_atapi) {
		/* check BIOSv2 boot mark -- remove it if you don't want this feature */
		fDeReadWrite(pDev, 9, IDE_COMMAND_READ, sector);		
		if (*(ULONG*)&sector[60]==LE32_TO_CPU(0x12345678) && sector[64])
			pVDev->vf_bootmark = 1;
	}
}
#endif

/********************************************************************
 * Initialization of a device table
 ********************************************************************/

void fSetDeviceTable(PDevice pDevice, PIDENTIFY_DATA pIdentify)
{
    PBadModeList      pbd;
    UCHAR             mod;

    pDevice->df_on_line = 1;

    if((pIdentify->SpecialFunctionsEnabled & LE16_TO_CPU(1)) || (pIdentify->GeneralConfiguration & LE16_TO_CPU(0x80))) 
    {
#ifdef SUPPORT_REMOVABLE_DISK
		pDevice->df_removable_drive = 1;
#else 
		if(pDevice->df_atapi) 
			pDevice->df_removable_drive = 1;
#endif
    }
	if(*(PULONG)pIdentify->ModelNumber == LE32_TO_CPU(0x2D314C53)) /* '-1LS') */
		pDevice->df_ls120 = 1;
 
    if (!pDevice->df_atapi && !pDevice->df_ls120)
        pDevice->df_harddisk = 1;
       
    if((pDevice->df_atapi) && memcmp((PUCHAR)&pIdentify->ModelNumber, (PUCHAR)"DCR-MO  DC", 10) == 0) 
    {
        pDevice->df_atapi_changer = 1;
        pDevice->bSubDevices = 3;
    }

/*===================================================================
 * Copy Basic Info
 *===================================================================*/   


    pDevice->df_lba = (pIdentify->Capabilities & LE16_TO_CPU(0x200))? 1 : 0;
   	/* 
	 *We can only set the maximum block to the half of the maximum disk cache buffer size 
	 *in order to get better perfermence
	 */
    pDevice->wWordsPerIntr = pIdentify->MaximumBlockTransfer << 8;

#ifdef SUPPORT_48BIT_LBA
    if(pIdentify->CommandSupport & LE16_TO_CPU(0x400)) {
		pDevice->df_support_48bit_lba = 1;
        pDevice->dDeRealCapacity = LE32_TO_CPU(pIdentify->Lba48BitLow) - 1;
		pDevice->bDeRealHeader = 255;
		pDevice->bDeRealSector = 63;
    }
    else
#endif
    if(pIdentify->TranslationFieldsValid & LE16_TO_CPU(1)) 
    {
		pDevice->bDeRealHeader     = (UCHAR)LE16_TO_CPU(pIdentify->NumberOfCurrentHeads);
       	pDevice->bDeRealSector     = (UCHAR)LE16_TO_CPU(pIdentify->CurrentSectorsPerTrack);
       	pDevice->dDeRealCapacity = ((LE32_TO_CPU(pIdentify->CurrentSectorCapacity) <
            LE32_TO_CPU(pIdentify->UserAddressableSectors))? LE32_TO_CPU(pIdentify->UserAddressableSectors) :
            LE32_TO_CPU(pIdentify->CurrentSectorCapacity)) - 1;
    } 
    else 
    {
		pDevice->bDeRealHeader     = (UCHAR)LE16_TO_CPU(pIdentify->NumberOfHeads);
       	pDevice->bDeRealSector     = (UCHAR)LE16_TO_CPU(pIdentify->SectorsPerTrack);
       	pDevice->dDeRealCapacity = LE32_TO_CPU(pIdentify->UserAddressableSectors) - 1;
	}

    pDevice->wDeRealHeadXsect = pDevice->bDeRealSector * pDevice->bDeRealHeader;

/*===================================================================
 * Select Best PIO mode
 *===================================================================*/   

    if((mod = pIdentify->PioCycleTimingMode) > 4)
        mod = 0;
    if((pIdentify->TranslationFieldsValid & LE16_TO_CPU(2)) && (pIdentify->Capabilities & LE16_TO_CPU(0x800)) && (pIdentify->AdvancedPIOModes)) 
    {
        mod = (UCHAR)(
            (pIdentify->AdvancedPIOModes & 0x2) ? 4 :
            (pIdentify->AdvancedPIOModes & 0x1) ? 3 : mod);
    }
    pDevice->bMaxShowMode = pDevice->bDeBestPIO = (UCHAR)mod;

/*===================================================================
 * Select Best Multiword DMA
 *===================================================================*/   

#ifndef NOT_SUPPORT_DMA
    if((pIdentify->Capabilities & LE16_TO_CPU(0x100)) && (pIdentify->MultiWordDMASupport & 6))  /* check mw dma */
	{
		pDevice->bDeBestDMA = (UCHAR)((pIdentify->MultiWordDMASupport & 4)? 2 : 1);
       	pDevice->bMaxShowMode = (UCHAR)(pDevice->bDeBestDMA + IDE_MWDMA_0);
    } 
    else 
#endif
		pDevice->bDeBestDMA = 0xFF;

	if (!IS_SATA_DEVICE(pDevice) && pDevice->pChannel->pChipInstance->ftbl.pfnCableDetect)
		pDevice->pChannel->pChipInstance->ftbl.pfnCableDetect(pDevice);

/*===================================================================
 * Select Best Ultra DMA
 *===================================================================*/   

#ifndef NOT_SUPPORT_DMA
    if(pIdentify->TranslationFieldsValid & LE16_TO_CPU(0x4))
    {
		mod=(UCHAR)((pIdentify->UtralDmaMode & LE16_TO_CPU(0x40))? 6 :
					(pIdentify->UtralDmaMode & LE16_TO_CPU(0x20))? 5 :
					(pIdentify->UtralDmaMode & LE16_TO_CPU(0x10))? 4 :
					(pIdentify->UtralDmaMode & LE16_TO_CPU(0x8)) ? 3 :
					(pIdentify->UtralDmaMode & LE16_TO_CPU(0x4)) ? 2 :    	/* ultra DMA Mode 2 */
					(pIdentify->UtralDmaMode & LE16_TO_CPU(0x2)) ? 1 : 0);   /* ultra DMA Mode 1 */

		if (pDevice->pChannel->cf_80pin_cable==0 && mod>2) mod = 2;
		
		pDevice->bDeBestUDMA = (UCHAR)mod;
        pDevice->bMaxShowMode = (UCHAR)(mod + IDE_UDMA_0);

	} 
	else
#endif
		pDevice->bDeBestUDMA = 0xFF;

/*===================================================================
 * select bset mode 
 *===================================================================*/   

    pbd = check_bad_disk((char *)&pIdentify->ModelNumber);

    if((pbd->UltraDMAMode | pDevice->bDeBestUDMA) != 0xFF) {
		pDevice->bDeUsable_Mode = (UCHAR)((MIN(pbd->UltraDMAMode, pDevice->bDeBestUDMA)) + 8);
#ifdef CLOCK_66MHZ
		if ((pDevice->pChannel->pChipInstance->ChipFlags & CHIP_SUPPORT_ULTRA_133)==0)
#endif
		{
			/* ATA/133 must work with 66MHz clock */
			if (pDevice->bDeUsable_Mode>13)
				pDevice->bMaxShowMode = pDevice->bDeUsable_Mode = 13;
		}
		if ((pDevice->pChannel->pChipInstance->ChipFlags & CHIP_DPLL_MODE)==0) {
			/* ATA/133 can't work when PCI clock less than 33MHz */
	    	extern int pci_clock_mhz;
			if (pci_clock_mhz<33 && pDevice->bDeUsable_Mode>13) 
				pDevice->bDeUsable_Mode = 13;
		}
	}
    else if((pbd->DMAMode | pDevice->bDeBestDMA) != 0xFF) 
		pDevice->bDeUsable_Mode = (UCHAR)((MIN(pbd->DMAMode, pDevice->bDeBestDMA)) + 5);
	else 
		pDevice->bDeUsable_Mode = MIN(pbd->PIOMode, pDevice->bDeBestPIO);


	if (!IS_SATA_DEVICE(pDevice)) {
		/* adjust Seagate Barracuda III/IV HDD settings */
		if (pIdentify->ModelNumber[0]==LE16_TO_CPU(0x5354)/*'ST'*/ && 
			pIdentify->ModelNumber[4]==LE16_TO_CPU(0x4120)/*'A '*/ &&
			(memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"132051", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"133501", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"230011", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"234041", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"336002", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"430061", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"630012", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"830012", 6)==0 ||
			 memcmp((PUCHAR)&pIdentify->ModelNumber[1], (PUCHAR)"438042", 6)==0))
		{
			void seagate_hdd_fix(PDevice pDev, PIDE_REGISTERS_1 IoPort, PIDE_REGISTERS_2 ControlPort);
			seagate_hdd_fix(pDevice, pDevice->pChannel->BaseIoAddress1, pDevice->pChannel->BaseIoAddress2);
		}
	}
}

/********************************************************************
 * check if the disk is "bad" disk
 ********************************************************************/
static BadModeList bad_disks[] = 
{
    {0xFF, 0xFF, 4, 8,  "TO-I79 5" },       	/* 0 */
    {3, 2, 4, 10,       "DW CCA6204" },     	/* 1 */
    {0xFF, 0xFF, 3, 10, "nIetrglaP " },     	/* 2 */
    {3, 2, 4, 10,       "DW CDW0000" },     	/* 3 reduce mode on AA series */
    {0xFF, 2, 4, 10,    "aMtxro9 01"},      	/* 4 reduce mode on 91xxxDx series */
    {0xFF, 2, 4, 14,    "aMtxro9 80544D"},  	/* 5 Maxtor 90845D4 */
    {0xFF, 0xFF, 4, 10, "eHlwte-taP"},      	/* 6 HP CD-Writer (0x5A cmd error) */

    {0xFF, 0xFF, 0, 16, "iPnoee rVD-DOR M"},	/* 8 PIONEER DVD-ROM */
    {0xFF, 0xFF, 4, 10, "DCR- WC XR" },     	/* 9 SONY CD-RW   (0x5A cmd error) */
    {0xFF, 0xFF, 0, 8,  "EN C    " },       	/* 10 */
    {0xFF, 1, 4, 18,    "UFIJST UPM3C60A5 H"}, 
    {0x2,  2, 4, 14,    "aMtxro9 80284U"},     	/* Maxtor 90882U4 */

    {3, 2, 4, 14,       "aMtxro9 02848U"},      /* Maxtor 92048U8 */
    {0x3,  2, 4, 16,   "UFIJST UPM3D60A4"},		/* Fujitsu MPD3064AT  */
	{0xFF, 0xFF, 4, 14, "ASSMNU GDCR-MO"},      /*SAMASUNG CD-ROM */
	/*add new here !! */

#ifdef CLOCK_66MHZ
	{ 6, 2, 4, 0, 0 }
#else 
	{ 5, 2, 4, 0, 0 }
#endif
};

#define MAX_BAD_DISKS (sizeof(bad_disks)/sizeof(BadModeList))


PBadModeList check_bad_disk(char *ModelNumber)
{
	int l;
	PBadModeList pbd;
	
	/*
	 * kick out the "bad device" which do not work with our chip
	 */
	for(l=0, pbd = bad_disks; l < MAX_BAD_DISKS - 1; pbd++,l++) {
		if(memcmp((PUCHAR)ModelNumber, (PUCHAR)pbd->name, pbd->length & 0x1F) == 0) {
			switch (l) {
			case 3:
				if(ModelNumber[3]== 'C' && ModelNumber[4]=='D' && ModelNumber[5]=='W' && 
					ModelNumber[8]=='A' && (ModelNumber[11]=='A' || 
					ModelNumber[10]=='A' || ModelNumber[9]=='A')) 
					goto out;
				break;
			case 4:
				if (*(DWORD*)ModelNumber==LE32_TO_CPU(0x78744d61) /*'xtMa'*/ &&
					ModelNumber[6]== '9' && ModelNumber[9]=='1' && ModelNumber[13] =='D')
					goto out;
				break;
			case 6:
				if (*(DWORD*)&ModelNumber[16]==LE32_TO_CPU(0x2d574344) /*'-WCD'*/ &&
					ModelNumber[20]== 'i' && ModelNumber[21]=='r' && ModelNumber[22] =='e')
					goto out;
				break;
			default:
				goto out;
			}
		}
	}
out:
	return(pbd);
}

void seagate_hdd_fix( PDevice pDev,
					PIDE_REGISTERS_1  IoPort,
					PIDE_REGISTERS_2  ControlPort
					)
{
	int i;

	mSetFeaturePort(IoPort, 0x00);		/* W 1F1 00 */
	mSetBlockCount(IoPort, 0x06);		/* W 1F2 06 */
	mSetBlockNumber(IoPort, 0x9A);		/* W 1F3 9A */
	mSetCylinderLow(IoPort, 0x00);		/* W 1F4 00 */
	mSetCylinderHigh(IoPort, 0x00);		/* W 1F5 00 */
	
	mSelectUnit(IoPort, pDev->bDeUnitId);	/* Select device */
	/*SelectUnit(IoPort, 0x00);			// W 1F6 00 */

	mIssueCommand(IoPort, 0x9A);			/* W 1F7 9A */

	fWaitOnBaseBusy(IoPort);				/* R 1F7 */

	mGetErrorCode(IoPort);				/* R 1F1 */
	mGetInterruptReason(IoPort);			/* R 1F2 */
	mGetUnitNumber(IoPort);		/* R 1F6 */
	mGetBlockNumber(IoPort);				/* R 1F3 */
	mGetByteLow(IoPort);					/* R 1F4 */
	mGetByteHigh(IoPort);				/* R 1F5 */
	mGetUnitNumber(IoPort);		/* R 1F6 */
	mGetBaseStatus(IoPort);				/* R 1F7 */
	mGetStatus(ControlPort);				/* R 3F6 */
	mGetBaseStatus(IoPort);				/* R 1F7 */

	/* */
	/* Write 512 bytes */
	/* */
	OutWord(&IoPort->Data, 0x5341);		/* W 1F0 5341 */
	OutWord(&IoPort->Data, 0x4943);		/* W 1F0 4943 */
	OutWord(&IoPort->Data, 0x3938);		/* W 1F0 3938 */
	OutWord(&IoPort->Data, 0x3831);		/* W 1F0 3831 */
	OutWord(&IoPort->Data, 0x3330);		/* W 1F0 3330 */
	OutWord(&IoPort->Data, 0x4646);		/* W 1F0 4646 */
	for(i = 0; i < 250; i++)
	{
		OutWord(&IoPort->Data, 0x0000);	/* W 1F0 0000 */
	}

	/* */
	/* Read 1F7 5 times */
	/* */
	for(i = 0; i < 5; i++)
	{
		mGetBaseStatus(IoPort);			/* R 1F7 */
	}
	__fWaitOnBusy(ControlPort);
}
