/*
 * ihpt3xx.c
 * initialize HPT3xx chips
 */
struct ChipProcTable hpt370_proc_table = {
	InitHpt3xxChip, 
	Hpt3xxModeSetting, 
	fHPT3xxCableDetect, 
	Hpt3xxControlIrq, 
	fHpt3xxFinishIntr,
	fHptReset370IdeEngine,
	fHpt3xxResetController,
	SetHpt3xxClk,
	fHpt3xxWaitOnBusy,
	TCQ_POLLING_FN(Hpt3xxControlTcqPolling)
	CLOCK_SWITCH_FN(fHPT370ClockCapable, fHPT370SwitchClock)
	PNP_MANAGEMENT_FN(Hpt3xxSetPnPConfig, 0, Hpt3xxSetRunningConfig, Hpt3xxHiberState)
};

struct ChipProcTable hpt3xx_proc_table = {
	InitHpt3xxChip, 
	Hpt3xxModeSetting, 
	fHPT3xxCableDetect, 
	Hpt3xxControlIrq, 
	fHpt3xxFinishIntr,
	fHptReset370IdeEngine,
	fHpt3xxResetController,
	SetHpt3xxClk,
	fHpt3xxWaitOnBusy,
	TCQ_POLLING_FN(Hpt3xxControlTcqPolling)
	CLOCK_SWITCH_FN(0, 0)
	PNP_MANAGEMENT_FN(Hpt3xxSetPnPConfig, 0, Hpt3xxSetRunningConfig, Hpt3xxHiberState)
};

struct ChipProcTable hpt3x2n_proc_table = {
	InitHpt3xxChip, 
	Hpt3xxModeSetting, 
	fHPT3xxCableDetect, 
	Hpt3xxControlIrq, 
	fHpt3xxFinishIntr,
	fHptReset370IdeEngine,
	fHpt3xxResetController,
	SetHpt3xxClk,
	fHpt3xxWaitOnBusy,
	TCQ_POLLING_FN(Hpt3xxControlTcqPolling)
	CLOCK_SWITCH_FN(fHPT3x2NClockCapable, fHPT3x2NSwitchClock)
	PNP_MANAGEMENT_FN(Hpt3xxSetPnPConfig, 0, Hpt3xxSetRunningConfig, Hpt3xxHiberState)
};

static int PCI_CLOCK_BY_FCNT(int fcnt, int default_dpll)
{
	int f = ((unsigned)fcnt*default_dpll/192);
	return
	 
		f<40? 33:
		f<45? 40:
		f<55? 50: 66;
}

/* can't use these vars after initilization! */
int dpll_freqency[] = { 48, 50, 66, 75 };

/******************************************************************
 * initialize Hpt3xx chips
 *******************************************************************/
void InitHpt3xxChip(PChipInstance pChip, UINT reset)
{
    PChannel       Primary, Secondry;
    PUCHAR         BMI;
    int            index;

    /* fix 0x70 before any initialization */
	pcicfg_write_byte(pChip->pci_bus, pChip->pci_dev, pChip->pci_func, 0x70, 0);

	BMI = pChip->pPciIoMap = (PUCHAR)(ULONG_PTR)(pcicfg_read_dword(pChip->pci_bus, 
		pChip->pci_dev, pChip->pci_func, 0x20) & ~1);
    /*
     * for all our chips according to INTEL spec.
     */
	pcicfg_write_byte(pChip->pci_bus, pChip->pci_dev, pChip->pci_func, REG_PCICMD, 5);

    OutPort(BMI + 0x20 + REG_MLT, 0x40);

	Hpt3xxIoEnable(pChip);

    /*
     * for HPT370,HPT370A,HPT370B
     */
    pChip->bCpMaxChannel = 2;
	if (!reset) {
		pChip->bChipID = (UCHAR)ChipChannels;
		ChipChannels += pChip->bCpMaxChannel;
	}

    pChip->ChipFlags = CHIP_SUPPORT_ULTRA_66 | CHIP_SUPPORT_ULTRA_100;
	if (pChip->bChipType > CHIP_TYPE_HPT370B)
		pChip->ChipFlags |= CHIP_SUPPORT_ULTRA_133;
	
	
	
		pChip->ChipFlags |= CHIP_LIMIT_U7;

	
	if (pChip->bChipType==CHIP_TYPE_HPT371 || pChip->bChipType==CHIP_TYPE_HPT371N)
    	OutPort(BMI + 0x20 + REG_MISC, 0);


	/* calculate PCI clock frequency */
	if (!pci_clock_mhz) {
		UINT f_cnt=0;
		ULONG total=0;
		/* first check if BIOS has saved f_CNT */
		total = InDWord(BMI+0x90);
		if ((total & 0xFFFFF000)==0xABCDE000)
			f_cnt = total & 0x1FF;
		if (f_cnt) {
			switch(pChip->bChipType) {
			case CHIP_TYPE_HPT372N:
			case CHIP_TYPE_HPT302N:
			case CHIP_TYPE_HPT371N:
				pci_clock_mhz = PCI_CLOCK_BY_FCNT(f_cnt, 77);
				break;
			case CHIP_TYPE_HPT302:
			case CHIP_TYPE_HPT371:
			case CHIP_TYPE_HPT372A:
				pci_clock_mhz = PCI_CLOCK_BY_FCNT(f_cnt, 66);
				break;
			case CHIP_TYPE_HPT372:
				pci_clock_mhz = PCI_CLOCK_BY_FCNT(f_cnt, 55);
				break;
			default:
				/* 370/374 */
				pci_clock_mhz = PCI_CLOCK_BY_FCNT(f_cnt, 48);
				break;
			}
		}
		else
			pci_clock_mhz = 33;

		KdPrint(("hpt3xx: detected %dMhz PCI clock", pci_clock_mhz));
	}

	SetHpt3xxClk(pChip);

    if (reset) {
		Primary = pChip->pChannel[0];
        Secondry = pChip->pChannel[1];
    }
    else
	{
        for(index = 0x60; index < 0x70; index += 4)
			OutDWord(BMI+index, pChip->Setting[0]);

        pChip->pChannel[0] = Primary = OsGetChannelTable(pChip, 0);
        pChip->pChannel[1] = Secondry = OsGetChannelTable(pChip, 1);

        Primary->BaseIoAddress1 = (PIDE_REGISTERS_1)(ULONG_PTR)(InDWord((PULONG)(BMI + 0x10)) & ~1);
        Primary->BaseIoAddress2 = (PIDE_REGISTERS_2)(ULONG_PTR)((InDWord((PULONG)(BMI + 0x14)) & ~1) + 2);

        Secondry->BaseIoAddress1 = (PIDE_REGISTERS_1)(ULONG_PTR)(InDWord((PULONG)(BMI + 0x18)) & ~1);
        Secondry->BaseIoAddress2 = (PIDE_REGISTERS_2)(ULONG_PTR)((InDWord((PULONG)(BMI + 0x1C)) & ~1) + 2);

		
		if (pci_clock_mhz<40 &&
			(pChip->bChipType==CHIP_TYPE_HPT372N || pChip->bChipType==CHIP_TYPE_HPT302N)) {
			pChip->pChannel[0]->bMaxMode = 14;
		}
    }

    Primary->cf_80pin_cable = Secondry->cf_80pin_cable = 1;
    Primary->BMI  = BMI;
    Secondry->BMI  = BMI + 8;
    Primary->MiscControlAddr = BMI + 0x20 + REG_MISC;
    Secondry->MiscControlAddr = BMI + 0x24 + REG_MISC;
    Primary->pChipInstance = Secondry->pChipInstance	= pChip;
}

void SetHpt3xxClk(PChipInstance pChip)
{
	PUCHAR BMI = pChip->pPciIoMap;
	int index, f_low;
#ifdef SUPPORT_DPLL
	int clk_index, f_high, adjust;
	ULONG loop;
#endif
	PChannel pChan;
	PDevice pDev;
	
	pChip->ChipFlags &= ~CHIP_HPT3XX_DPLL_MODE;
	
#ifdef SUPPORT_DPLL
	/* 
	 * adjust DPLL clock. 
	 * If there is ATA/150 disk, use DPLL75, else
	 * if there is ATA/133 disk, use DPLL66, else use DPLL50.
	 */
	f_high = 0;
	for (index=0; index<pChip->bCpMaxChannel; index++) {
		pChan = pChip->pChannel[index];
		if (pChan) for (f_low=0; f_low<2; f_low++) {
			pDev = pChan->pDevice[f_low];
			if (pDev && pDev->bDeModeSetting>f_high)
				f_high = pDev->bDeModeSetting;
		}
	}
	if (f_high>14 && (pChip->ChipFlags & CHIP_LIMIT_U7)==0)
		clk_index = 3;
	else if (f_high>13)
		clk_index = 2;
	else
		clk_index = 1;

	f_low = pci_clock_mhz*48/dpll_freqency[clk_index] - 1;
	f_high = f_low + 2;
	adjust = 0;

reset_5C:
	OutDWord(BMI+0x7C, ((ULONG)f_high << 16) | f_low | 0x100);
    OutPort(BMI + 0x7B, 0x21);
wait_stable:
	for(loop = 0; loop < 0x5000; loop++) {
		StallExec(50);
		if (InPort(BMI + 0x7B) & 0x80) {
			for(loop = 0; loop < 0x1000; loop++)
				if((InPort(BMI + 0x7B) & 0x80) == 0) goto re_try;
			OutDWord(BMI+0x7C, InDWord(BMI+0x7C) & ~0x100);
           	goto dpll_ok;
		}
	}
re_try:
	if(++adjust < 8) {
		if (0) { /*((adjust & 3)==3) { */
			f_low--; f_high++; 
			goto  reset_5C;
		}
		else {
			OutDWord(BMI+0x7C, ((ULONG)f_high << 16) | f_low);
			goto wait_stable;
		}
	}
	KdPrint(("hpt3xx: can't wait DPLL stable, force 0x%lx", ((ULONG)f_high << 16) | f_low));
dpll_ok:
	pChip->ChipFlags |= CHIP_HPT3XX_DPLL_MODE;
	pChip->Setting = 
#ifdef CLOCK_66MHZ
					clk_index==3? setting370_75 :
					clk_index==2? setting370_66 : 
#endif
					setting370_50;
	pChip->using_dpll = 1;

#endif
	
	if ((pChip->ChipFlags & CHIP_HPT3XX_DPLL_MODE)==0) {
		/* use PCI clock */
    	OutPort(BMI + 0x7B, 0x23);
    	pChip->Setting = setting370_33;
    }
    /* set device timing registers */
	for (index=0; index<pChip->bCpMaxChannel; index++) {
		pChan = pChip->pChannel[index];
		if (pChan) for (f_low=0; f_low<2; f_low++) {
			pDev = pChan->pDevice[f_low];
			if (pDev) {
				OutDWord((PULONG)(pChan->BMI + ((pDev->bDeUnitId & 0x10)>>2) + 0x60),
					pChip->Setting[pDev->bDeModeSetting]);
			}
		}
	}
}
