/***************************************************************************
 * File:         init.c
 * Description:  HPT3xx driver initialization
 * Copyright (c) 2000-2003 HighPoint Technologies, Inc. All rights reserved.
 ***************************************************************************/

#include "global.h"


int InitializeAllChips(void)
{
    PChipInstance pChip = gChipInstances;
#ifdef SUPPORT_ARRAY
    int nDev;
	int j;
#endif

    ZeroMemory(gVBus, sizeof(VBus)*MAX_VBUS);

#ifdef SUPPORT_ARRAY
	for (j=0; j<MAX_VBUS; j++) {
		_VBUS_INST(&gVBus[j]);
		for(nDev = MAX_ARRAY_PER_VBUS - 1; nDev >= 0; nDev--) {
			PVDevice pVArray = ArrayTables(nDev);
			mArFreeArrayTable(pVArray);
		}
	}
#endif

	KdPrint(("Scan for chips..."));
    /*
     * Scan Chip
     */
    for ( ; ; ) 
	{
		ZeroMemory(pChip, sizeof(ChipInstance));
		if(GetNextPciInfo(pChip) == 0)
            break;
        pChip++;
		if (pChip-gChipInstances>=MAX_CHIPS) break;
    }

    if(pChip == gChipInstances) {
    	KdPrint(("No supported chip found"));
        return(0);
	}


    /*
     * Initial all chips and the channels attached on them
     */
    pChanStart = (PChannel)pChip;
    pDataPool = (char *)pChip;

    for(pChip = gChipInstances; pChip < pChipEnd; pChip++) 
		pChip->ftbl.pfnInitialize(pChip, INIT_ENTRY);    
	
	for(pChip = gChipInstances; pChip < pChipEnd; pChip++)
		pChip->ftbl.pfnControlIrq(pChip, 0);
		
    pVDevStart = (PVDevice)pDataPool;
	return 0;
}

void InitializeVBus(PVBus pVBus)
{
	_VBUS_INST(pVBus)
    PChipInstance pChip;
	PVDevice pVDev;
	UINT j;

	KdPrint(("Scan for devices..."));
    /*
     * scan all devices on each channel
     */
    for(j=0; j<pVBus->nInstances; j++) {
    	pChip = pVBus->pChipInstance[j];
        fDeScanAllDevices(pChip);
        /* 
		 * check ATA/133 device to adjust DPLL clock. 
		 */
		if (pChip->ftbl.pfnInit2) pChip->ftbl.pfnInit2(pChip);
    }

	
	for (pVDev=pVDevStart; pVDev<pVDevEnd; pVDev++) {
		if (pVDev->pVBus==pVBus &&
			pVDev->VDeviceType==VD_SINGLE_DISK &&
			pVDev->pParent==0)
			fCheckBootable(pVDev);
	}

#ifdef SUPPORT_ARRAY
	CheckArrayCritical(_VBUS_P0);
	
	
	for (j=0; j<MAX_ARRAY_PER_VBUS; j++) {
		pVDev = ArrayTables(j);
		if (pVDev->u.array.dArStamp==0 || pVDev->pParent) continue;
		fCheckBootable(pVDev);
	}
#endif

	/* register all logical devices on vbus */
	__fRegisterVDevices(pVBus);

	/* enable interrupt */
    for(j=0; j<pVBus->nInstances; j++) {
    	pChip = pVBus->pChipInstance[j];
		pChip->ftbl.pfnControlIrq(pChip, 1);
	}


}

int Initialize(void)
{
	int j;
	InitializeAllChips();
	for (j=0; j<MAX_VBUS && gVBus[j].nInstances; j++)
		InitializeVBus(&gVBus[j]);
	return pVDevEnd - pVDevStart;
}

PChannel fGetChannelTable(void)
{    
     PChannel pChan = (PChannel)pDataPool;
     pDataPool += sizeof(Channel);
     ZeroMemory((char *)pChan,sizeof(Channel));
     return(pChan);
}

PDevice  fGetDeviceTable(void)
{
    PVDevice pVDev;
    
    for (pVDev=pVDevStart; pVDev<pVDevEnd; pVDev++)
    	if (pVDev->vf_freed) break;
    pDataPool += sizeof(VDevice);
    ZeroMemory((char *)pVDev,sizeof(VDevice));
    return &pVDev->u.disk;
}

void  OsReturnTable(PDevice pDevice)
{
    pDataPool = (char *)Map2pVDevice(pDevice); 
}

