/*
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * Copyright (C) 2000-2003 HighPoint Technologies, Inc.
 * oslinux.h  Linux headers for platform dependent part.
 */
#ifndef _OSLINUX_H_
#define _OSLINUX_H_

#define hpt_printk(x) do { printk(PROC_DIR_NAME": "); printk x; } while (0)

/*************************************************************************
 * hardware I/O functions                                                *
 *************************************************************************/
#define T_IOPORT(x) ((USHORT)(ULONG_PTR)(x))
#define OutDWord(x, y)    outl(y, T_IOPORT(x))
#define InDWord(x)        inl(T_IOPORT(x))
#define OutPort(x, y)     outb(y, T_IOPORT(x))
#define InPort(x)         inb(T_IOPORT(x))
#define OutWord(x, y)     outw(y, T_IOPORT(x))
#define InWord(x)         inw(T_IOPORT(x))
#define RepINS(x,y,z)     insw(T_IOPORT(&x->Data), (PUSHORT)y, z)
#define RepOUTS(x,y,z)    outsw(T_IOPORT(&x->Data), (PUSHORT)y, z)
#define RepINSD(x,y,z)    insd(T_IOPORT(&x->Data), (PULONG)y, z)
#define RepOUTSD(x,y,z)   outsd(T_IOPORT(&x->Data), (PULONG)y, z)

/*************************************************************************
 * PCI configuration space access
 *************************************************************************/
UCHAR pcicfg_read_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg);
WORD  pcicfg_read_word (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg);
DWORD pcicfg_read_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg);
void pcicfg_write_byte (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, UCHAR v);
void pcicfg_write_word (UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, WORD v);
void pcicfg_write_dword(UCHAR bus, UCHAR dev, UCHAR func, UCHAR reg, DWORD v);


extern struct Scsi_Host *hpt_vhost;

typedef struct 
{
	UCHAR		status;     		/* 0 nonbootable; 80h bootable */
	UCHAR      	start_head;
	USHORT     	start_sector;
	UCHAR      	type;
	UCHAR      	end_head;
	USHORT     	end_sector;
	ULONG      	start_abs_sector;
	ULONG      	num_of_sector;
} partition;

typedef struct _INQUIRYDATA {
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceType : 5;
	UCHAR RemovableMedia : 1;
	UCHAR DeviceTypeModifier : 7;
#else 
	UCHAR DeviceType : 5;
	UCHAR DeviceTypeQualifier : 3;
	UCHAR DeviceTypeModifier : 7;
	UCHAR RemovableMedia : 1;
#endif
	UCHAR Versions;
	UCHAR ResponseDataFormat;
	UCHAR AdditionalLength;
	UCHAR Reserved[2];
#ifdef __BIG_ENDIAN_BITFIELD
	UCHAR RelativeAddressing : 1;
	UCHAR Wide32Bit : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Synchronous : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Reserved2 : 1;
	UCHAR CommandQueue : 1;
	UCHAR SoftReset : 1;
#else 
	UCHAR SoftReset : 1;
	UCHAR CommandQueue : 1;
	UCHAR Reserved2 : 1;
	UCHAR LinkedCommands : 1;
	UCHAR Synchronous : 1;
	UCHAR Wide16Bit : 1;
	UCHAR Wide32Bit : 1;
	UCHAR RelativeAddressing : 1;
#endif
	UCHAR VendorId[8];
	UCHAR ProductId[16];
	UCHAR ProductRevisionLevel[4];
	UCHAR VendorSpecific[20];
	UCHAR Reserved3[40];
} INQUIRYDATA, *PINQUIRYDATA;

struct hpt3xx_HOSTDATA {
	Scsi_Cmnd *pending_Q;
		/* pending commands to issue */
	Scsi_Cmnd *current_Q;
		/* current commands on the bus */
	UCHAR commands;		
	    /* number of current commands,this way only allow commands==32 */
	UCHAR status;
		/* host lock */
	UCHAR irq[MAX_CHIPS];
	struct pci_dev *pci_dev;
};
/* access macros for hostdata */
#define HOSTDATA(shpnt)		((struct hpt3xx_HOSTDATA *) &hpt_vhost->hostdata)
#define PENDING_Q           (HOSTDATA(shpnt)->pending_Q)
#define CURRENT_Q		    (HOSTDATA(shpnt)->current_Q)
#define CURRCMDs            (HOSTDATA(shpnt)->commands)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define sc_host(sc)    sc->device->host
#define sc_channel(sc) sc->device->channel
#define sc_target(sc)  sc->device->id
#define sc_lun(sc)     sc->device->lun
#define sc_flags(sc)     sc->eh_eflags
#else 
#define sc_host(sc)    sc->host
#define sc_channel(sc) sc->channel
#define sc_target(sc)  sc->target
#define sc_lun(sc)     sc->lun
#define sc_flags(sc)   sc->flags
#endif

struct hpt_scsi_pointer {
	char * ptr;
	int status;
	int mapped;
	int sgcnt;
	dma_addr_t dma_handle;
};

#define HPT_SCP(SCpnt) ((struct hpt_scsi_pointer *)&(SCpnt)->SCp)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define pci_map_sg(pdev, sg_list, nseg, direction) (nseg)
#define pci_unmap_sg(pdev, sg_list, nseg, direction)
#define pci_map_single(pdev,buffer,size,direction) virt_to_bus(buffer)
#define pci_unmap_single(pdev,bus_addr,size,direction)
#define sg_dma_address(sg) (virt_to_bus((sg)->address))
#define sg_dma_len(sg) ((sg)->length)
#endif

extern spinlock_t driver_lock;

/*entry.c*/
int HPTLIBAPI fOsBuildSgl(PCommand pCmd, FPSCAT_GATH pSg, int logical);
typedef void (*HPT_DPC)(void*,UCHAR);
int hpt_queue_dpc(HPT_DPC dpc, void *arg, UCHAR flags);
void hpt_rebuild_data_block(PVDevice pArray, UCHAR flags);

#define DUPLICATE      0
#define INITIALIZE     1
#define REBUILD_PARITY 2
#define VERIFY_R         3

#endif
