/***************************************************************************
 * File:         vdevice.c
 * Description:  virtual device management
 * Copyright (c) 2000-2003 HighPoint Technologies, Inc. All rights reserved.
 ***************************************************************************/

#include "global.h"

/***************************************************************************
 * Function:     void fRegisterChip(PChipInstance pChip)
 * Description:  Regist a chip to a VBus
 * Argument:     PChipInstance pChip -- The Chip need to be registered
 * Retures:      None
 *               
 ***************************************************************************/

void fRegisterChip(PChipInstance pChip)
{
	PVBus  pVBus;

	FOR_EACH_VBUS(pVBus) {
		if (pVBus->nInstances==0) goto register_chip;
		if (CHIP_ON_SAME_VBUS(pChip, pVBus->pChipInstance[0])) goto register_chip;
	}
	return;

register_chip:
	if(pVBus->nInstances < MAX_CHIP_IN_VBUS) {
		pVBus->pChipInstance[pVBus->nInstances++] = pChip;
		pChip->pVBus = pVBus;
	}
}

/***************************************************************************
 * Function:     void  fRegisterVDevices(void);
 * Description:  Register all virtural devices to vbus
 * Retures:      None
 ***************************************************************************/
PVDevice fGetFirstChild(PVDevice pLogical)
{
#ifdef SUPPORT_ARRAY
	int i;
	if (mIsArray(pLogical)) {
		for (i=0; i<pLogical->u.array.bArnMember; i++)
			if (pLogical->u.array.pMember[i])
				return fGetFirstChild(pLogical->u.array.pMember[i]);
		return 0;
	}
#endif
	return pLogical;
}

void fResetBootMark(PVDevice pLogical)
{
#ifdef SUPPORT_ARRAY
	int i;
	if (mIsArray(pLogical)) {
		for (i=0; i<pLogical->u.array.bArnMember; i++)
			if (pLogical->u.array.pMember[i])
				fResetBootMark(pLogical->u.array.pMember[i]);
	}
#endif
	pLogical->vf_bootmark = 0;
}

void __fRegisterVDevices(PVBus pVBus)
{
	PVDevice pPhysical, pLogical;
	int i;

	ZeroMemory(pVBus->pVDevice, sizeof(pVBus->pVDevice));
	for(pPhysical = pVDevStart; pPhysical < pVDevEnd; pPhysical++) {
		if (pPhysical->pVBus!=pVBus) continue;
		pLogical = pPhysical;
		while (pLogical->pParent) pLogical = pLogical->pParent;
		if (pLogical->vf_online==0) {
			pPhysical->vf_bootmark = pLogical->vf_bootmark = 0;
			continue;
		}
		if (pLogical->VDeviceType==VD_SPARE || pPhysical!=fGetFirstChild(pLogical))
			continue;

		i=0;
		while (i<MAX_VDEVICE_PER_VBUS && pVBus->pVDevice[i]) i++;
		if (i<MAX_VDEVICE_PER_VBUS) {

			pVBus->pVDevice[i] = pLogical;
			pLogical->pVBus = pVBus;

			KdPrint(("registerVDev: pLogical=%p bootmark=%d bootable=%d",
				pLogical, pLogical->vf_bootmark, pLogical->vf_bootable));

			/* If this device is boot disk then assign to pVDevice[0] */
			if (i>0 && pLogical->vf_bootmark) {
				if (pVBus->pVDevice[0]->vf_bootmark) {
					
					if (pVBus->pVDevice[0]->vf_bootable || !pLogical->vf_bootable)
						fResetBootMark(pLogical);
					else {
						fResetBootMark(pVBus->pVDevice[0]);
						pVBus->pVDevice[i] = pVBus->pVDevice[0];
						pVBus->pVDevice[0] = pLogical;
					}
				}
				else {
					do { pVBus->pVDevice[i] = pVBus->pVDevice[i-1]; } while (--i);
					pVBus->pVDevice[0] = pLogical;
				}
			}
		}
	}
}

void fRegisterVDevices(void)
{
	PVBus  pVBus;

	FOR_EACH_VBUS(pVBus) {
		__fRegisterVDevices(pVBus);
	}
}


/*
 * Unregister a device from VBus.
 * The caller, if possible, shall unregister the device from OS before calling
 * this function.
 */
void HPTLIBAPI UnregisterVDevice(PVDevice pVDev)
{
	PVBus  pVBus = pVDev->pVBus;
	UINT   i;

	for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++)
		if(pVBus->pVDevice[i] == pVDev) {
			pVBus->pVDevice[i] = 0;
			break;
		}
}

void HPTLIBAPI fOsDiskFailed(_VBUS_ARG PVDevice pVDev)
{
	pVDev->vf_online = 0;
	/* force OS to unregister it */
	OsUnregisterDevice(pVDev);
	UnregisterVDevice(pVDev);
}

void HPTLIBAPI fCheckBootable(PVDevice pVDev)
{		
	int	i;
	DECLARE_BUFFER(PMaster_Boot_Record, mbr);
	PVDevice pDisk = pVDev;
	
	if(!pVDev->vf_online) return; /* this device is disabled */
	
	if(pVDev->VDeviceType < VD_SINGLE_DISK)	return;
	pVDev->vf_bootable = 0;
	ZeroMemory(mbr, sizeof(Master_Boot_Record));
			
#ifdef SUPPORT_ARRAY
	while(pDisk->VDeviceType != VD_SINGLE_DISK) {
		pDisk = pDisk->u.array.pMember[0];
		if (pVDev->VDeviceType == VD_RAID_5) {
			if (pDisk==NULL || pVDev->u.array.CriticalMembers==1) {
				int k;
				DECLARE_BUFFER(PMaster_Boot_Record, mbr_tmp); /* Take care in BIOS! */
				for( i = 1; i < pVDev->u.array.bArnMember; i++)	{				
					PVDevice pMember = pVDev->u.array.pMember[i];
					HPT_ASSERT(pMember);
					fDeReadWrite(&pMember->u.disk, 0, IDE_COMMAND_READ, mbr_tmp);
					/* Xor mbr_tmp to mbr. DoXor2 is not available if SUPPORT_CACHE undefined */
					for (k=0; k<128; k++) ((ULONG*)mbr)[k] ^= ((ULONG *)mbr_tmp)[k];
				}
				goto check_mbr;
			}
		}
		else {
			HPT_ASSERT(pDisk);
		}
	}
#endif
	
	fDeReadWrite(&pDisk->u.disk, 0, IDE_COMMAND_READ, mbr);

#ifdef SUPPORT_ARRAY
check_mbr:
#endif
	if(mbr->signature==CPU_TO_LE16(0xAA55)) 
	{	
		/* Some linux version will not set bootid to 0x80. Check "LILO" or "GRUB" */
		if (mbr->parts[0].numsect && 
			(*(ULONG*)&mbr->bootinst[6]==CPU_TO_LE32(0x4F4C494C) || *(ULONG*)&mbr->bootinst[374]==CPU_TO_LE32(0x42555247)))
			pVDev->vf_bootable = 1;
		else
			for (i=0; i<4; i++) 		
				if (mbr->parts[i].bootid==0x80 && mbr->parts[i].numsect) 								
					pVDev->vf_bootable = 1;
	}			
}

/***************************************************************************
 * defered procedure call
 ***************************************************************************/
void FASTCALL CallAfterReturn(_VBUS_ARG DPC_PROC proc, void *arg)
{
	int p = (_vbus_(PendingRoutinesLast) + 1) % MAX_PENDING_ROUTINES;
	HPT_ASSERT(p!=_vbus_(PendingRoutinesFirst));
	_vbus_(PendingRoutines)[_vbus_(PendingRoutinesLast)].proc = proc;
	_vbus_(PendingRoutines)[_vbus_(PendingRoutinesLast)].arg = arg;
	_vbus_(PendingRoutinesLast) = p;
}
void HPTLIBAPI CheckPendingCall(_VBUS_ARG0)
{
	while (_vbus_(PendingRoutinesFirst)!=_vbus_(PendingRoutinesLast)) {
		DPC_ROUTINE p = _vbus_(PendingRoutines)[_vbus_(PendingRoutinesFirst)];
		_vbus_(PendingRoutinesFirst)++;
		_vbus_(PendingRoutinesFirst) %= MAX_PENDING_ROUTINES;
		p.proc(_VBUS_P p.arg);
	}
}
void FASTCALL CallWhenIdle(_VBUS_ARG DPC_PROC proc, void *arg)
{
	int p = (_vbus_(IdleRoutinesLast) + 1) % MAX_IDLE_ROUTINES;
	HPT_ASSERT(p!=_vbus_(IdleRoutinesFirst));
	HPT_ASSERT(proc);
	_vbus_(IdleRoutines)[_vbus_(IdleRoutinesLast)].proc = proc;
	_vbus_(IdleRoutines)[_vbus_(IdleRoutinesLast)].arg = arg;
	_vbus_(IdleRoutinesLast) = p;
}
void HPTLIBAPI CheckIdleCall(_VBUS_ARG0)
{
	while (_vbus_(IdleRoutinesFirst)!=_vbus_(IdleRoutinesLast)) {
		DPC_ROUTINE p = _vbus_(IdleRoutines)[_vbus_(IdleRoutinesFirst)];
		_vbus_(IdleRoutinesFirst)++;
		_vbus_(IdleRoutinesFirst) %= MAX_IDLE_ROUTINES;
#if defined(SUPPORT_ARRAY) && defined(_RAID5N_)
		
		if (_vbus_(r5.active_stripes)) {
			PVBus _vbus_p = gVBus;
			fCompleteAllCommandsSynchronously(_vbus_p);
		}
#endif
		p.proc(_VBUS_P p.arg);
	}
}

/***************************************************************************
 * command queue
 ***************************************************************************/
PCommand HPTLIBAPI AllocateCommand(_VBUS_ARG0)
{
	if (_vbus_(pFreeCommands)) {
		PCommand ret = _vbus_(pFreeCommands);
		_vbus_(pFreeCommands) = *(PCommand *)_vbus_(pFreeCommands);
		ZeroMemory(ret, sizeof(Command));
		return ret;
	}
	return 0;
}
void FASTCALL FreeCommand(_VBUS_ARG PCommand pCmd)
{
	*(PCommand *)pCmd = _vbus_(pFreeCommands);
	_vbus_(pFreeCommands) = pCmd;
}
PCommand FASTCALL GetCommandFromQueue(PDevice pDev)
{
	PCommand pCmd=0;
	if (pDev->QueuedJob) {
		pCmd = pDev->QueuedJob->uScratch.wait.pNext;
		if (pDev->QueuedJob==pCmd)
			pDev->QueuedJob = NULL;
		else
			pDev->QueuedJob->uScratch.wait.pNext = pCmd->uScratch.wait.pNext;
		pDev->QueueLength--;
	}
	return pCmd;
}
PCommand FASTCALL PickCommandFromQueue(PDevice pDev)
{
	if (pDev->QueuedJob)
		return pDev->QueuedJob->uScratch.wait.pNext;
	return 0;
}
BOOLEAN FASTCALL PutCommandToQueue(PDevice pDev, PCommand pCmd)
{
	AddToWaitingList(&pDev->QueuedJob, pCmd);
	pDev->QueueLength++;
	return 1;
}
void HPTLIBAPI AddToWaitingList(PCommand *ppList, PCommand pCmd)
{
	if (*ppList) {
		pCmd->uScratch.wait.pNext = (*ppList)->uScratch.wait.pNext;
		(*ppList)->uScratch.wait.pNext = pCmd;
	}
	else
		pCmd->uScratch.wait.pNext = pCmd;
	*ppList = pCmd;
}
void HPTLIBAPI DoWaitingList(_VBUS_ARG PCommand *ppList)
{
	PCommand pCmd, pLast;
	if (*ppList) {
		KdPrint(("Check waiting list %x", ppList));
		pLast = *ppList;
		do {
			pCmd = (*ppList)->uScratch.wait.pNext;
			if (*ppList==pCmd)
				*ppList = NULL;
			else
				(*ppList)->uScratch.wait.pNext = pCmd->uScratch.wait.pNext;
			KdPrint(("launch cmd for %x Lba %x Sectors %x", 
				pCmd, pCmd->uCmd.Ide.Lba, pCmd->uCmd.Ide.nSectors));
			/* pCmd->uScratch.wait.WaitEntry may call DoWaitingList */
			pCmd->uScratch.wait.WaitEntry(_VBUS_P pCmd);
		}
		while (pCmd!=pLast && *ppList);
		KdPrint(("Waiting list became %x", *ppList));
	}
}


/*
 * flush a logical device.
 */
void HPTLIBAPI fFlushVDev(PVDevice pVDev)
{
#if defined(SUPPORT_ARRAY) && defined(_RAID5N_)
	if (pVDev->VDeviceType==VD_RAID_5
		) {
		flush_raid5(pVDev);
	}
#endif
}


void HPTLIBAPI fShutdownVDev(PVDevice pVDev)
{
	fFlushVDev(pVDev);
#ifdef SUPPORT_ARRAY
	if (mIsArray(pVDev)) {
		UINT i;
		for (i=0; i<pVDev->u.array.bArnMember; i++) {
			PVDevice pMember = pVDev->u.array.pMember[i];
			if (pMember && pMember->vf_online)
				fShutdownVDev(pMember);
		}
	}
	else 
#endif
	if (pVDev->u.disk.df_harddisk && pVDev->u.disk.df_on_line) {
		fNonIoAtaCmd(&pVDev->u.disk, IDE_COMMAND_FLUSH_CACHE);
	}
}

static void fResetActiveCommands(PVBus _vbus_p)
{
	PChannel pChan;
	PCommand pCmd;

	/* hardware reset. */
	FOR_EACH_CHANNEL_ON_VBUS(_vbus_p, pChan) {
		pChan->cf_resetting = 1;
		if (pChan->pChWorkDev
		){
			if (pChan->pChipInstance->ftbl.pfnResetController)
				pChan->pChipInstance->ftbl.pfnResetController(pChan);
			/* abort the current active commands */
			if (pChan->pChWorkDev) {
				pCmd = pChan->CurrentCmd;
				HPT_ASSERT(pCmd);
				pCmd->Result = RETURN_BUS_RESET;
				CallAfterReturn(_VBUS_P (DPC_PROC)pCmd->pfnCompletion, pCmd);
				
			}
		}
		else
			HPT_ASSERT(pChan->CurrentCmd==0);
		pChan->cf_resetting = 0;
	}
}

void HPTLIBAPI fCompleteAllCommandsSynchronously(PVBus _vbus_p)
{
	PChipInstance pCI;
	UINT i, j, cont;
	ULONG ticks = 0;
	do {
check_cmds:
		cont = 0;
		CheckPendingCall(_VBUS_P0);
#ifdef _RAID5N_
		dataxfer_poll();
		xor_poll();
#endif
		for(i = 0; i < _vbus_p->nInstances; i++) {
			pCI = _vbus_p->pChipInstance[i];
			for (j=0; j<pCI->bCpMaxChannel; j++) {
				if (pCI->pChannel[j]->CurrentCmd) {
					if (!fChipInterrupt(pCI)) {
						StallExec(1000);
						ticks++;
						if (ticks>3000) {
							fResetActiveCommands(_vbus_p);
							ticks = 0;
							goto check_cmds;
						}
					}
					else
						ticks = 0;
					cont = 1;
					/* check next chip */
					break;
				}
			}
		}
	}
	while (cont);
}

/* 
 * reset controller routine.
 *  Finish all pending commands.
 *  Context requirements:
 *    1. Synchronizing fResetBus with fOsSendCommand() and fChipInterrupt().
 *    2. Can't call fResetBus in a completion routine.
 *  All commands' completion routine will be called before fResetBus() returns.
 */
void HPTLIBAPI fResetVBus(_VBUS_ARG0)
{
	PVBus _vbus_p = gVBus;
	PChannel pChan;

	KdPrint(("fResetVBus(%d)", _vbus_p-gVBus));

	/* some commands may already finished. */
	CheckPendingCall(_VBUS_P0);

	fResetActiveCommands(_vbus_p);
	/* 
	 * the other pending commands may still be finished successfully.
	 */
	fCompleteAllCommandsSynchronously(_vbus_p);

	/* Now there should be no pending commands. No more action needed. */
	FOR_EACH_CHANNEL_ON_VBUS(_vbus_p, pChan) {
		HPT_ASSERT(pChan->CurrentCmd==0 && pChan->pChWorkDev==0);
		HPT_ASSERT(pChan->pDevice[0]==0 || pChan->pDevice[0]->QueuedJob==0);
		HPT_ASSERT(pChan->pDevice[1]==0 || pChan->pDevice[1]->QueuedJob==0);
#ifdef CHECK_HARDWARE_TIMEOUT
		HPT_ASSERT(pChan->Timer==0);
#endif

	}


	CheckIdleCall(_VBUS_P0);

	KdPrint(("fResetVBus() done"));
}

